# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, TypeVar

from nncf.common.deprecation import warning_deprecated
from nncf.common.quantization.structs import QuantizationPreset
from nncf.parameters import ModelType
from nncf.parameters import QuantizationMode
from nncf.parameters import TargetDevice
from nncf.quantization.advanced_parameters import AdvancedQuantizationParameters
from nncf.quantization.algorithms.bias_correction.algorithm import BIAS_CORRECTION_THRESHOLD
from nncf.quantization.algorithms.bias_correction.algorithm import BiasCorrection
from nncf.quantization.algorithms.channel_alignment.algorithm import ChannelAlignment
from nncf.quantization.algorithms.fast_bias_correction.algorithm import FAST_BIAS_CORRECTION_THRESHOLD
from nncf.quantization.algorithms.fast_bias_correction.algorithm import FastBiasCorrection
from nncf.quantization.algorithms.min_max.algorithm import MinMaxQuantization
from nncf.quantization.algorithms.pipeline import Pipeline
from nncf.quantization.algorithms.smooth_quant.algorithm import SmoothQuant
from nncf.scopes import IgnoredScope

TModel = TypeVar("TModel")


def create_ptq_pipeline(
    mode: Optional[QuantizationMode] = None,
    preset: Optional[QuantizationPreset] = None,
    target_device: TargetDevice = TargetDevice.ANY,
    subset_size: int = 300,
    fast_bias_correction: bool = True,
    model_type: Optional[ModelType] = None,
    ignored_scope: Optional[IgnoredScope] = None,
    advanced_parameters: Optional[AdvancedQuantizationParameters] = None,
) -> Pipeline:
    """
    Creates a post-training quantization pipeline.

    The post-training quantization pipeline includes the following steps:
        1) SmoothQuant
        2) ChannelAlignment
        3) MinMaxQuantization
        4) FastBiasCorrection or BiasCorrection

    :param mode: Special quantization mode that specify different ways of the optimization.
    :param preset: A preset controls the quantization mode (symmetric and asymmetric).
        It can take the following values:
        - `performance`: Symmetric quantization of weights and activations.
        - `mixed`: Symmetric quantization of weights and asymmetric quantization of activations.
        Default value is None. In this case, `mixed` preset is used for `transformer`
        model type otherwise `performace`.
    :param target_device: A target device the specificity of which will be taken
        into account while compressing in order to obtain the best performance
        for this type of device.
    :param subset_size: Size of a subset to calculate activations
        statistics used for quantization.
    :param fast_bias_correction: Setting this option to `False` enables a different
        bias correction method which is more accurate, in general, and takes
        more time but requires less memory.
    :param model_type: Model type is needed to specify additional patterns
        in the model. Supported only `transformer` now.
    :param ignored_scope: An ignored scope that defined the list of model control
        flow graph nodes to be ignored during quantization.
    :param advanced_parameters: Advanced quantization parameters for
        fine-tuning the quantization algorithm
    :return: A post-training quantization pipeline.
    """

    if advanced_parameters is None:
        advanced_parameters = AdvancedQuantizationParameters()

    # Build the post-training quantization pipeline.
    pipeline_steps = []

    # Add the `SmoothQuant` algorithm as the first step of the pipeline.
    # It is added only for `ModelType.TRANSFORMER`.
    sq_params = advanced_parameters.smooth_quant_alphas
    sq_alpha = advanced_parameters.smooth_quant_alpha
    if sq_alpha is not None:
        warning_deprecated(
            "`AdvancedQuantizationParameters(smooth_quant_alpha=..)` is deprecated."
            "Please, use `AdvancedQuantizationParameters(smooth_quant_alphas)` option "
            "with AdvancedSmoothQuantParameters(convolution=.., matmul=..) as value instead."
        )
        if sq_alpha < 0:
            sq_params.convolution = -1
            sq_params.matmul = -1
        else:
            sq_params.matmul = sq_alpha

    if model_type == ModelType.TRANSFORMER and (sq_params.convolution >= 0 or sq_params.matmul >= 0):
        alpha_map = {"convolution": sq_params.convolution, "matmul": sq_params.matmul}
        pipeline_steps.append([SmoothQuant(subset_size, advanced_parameters.inplace_statistics, alpha_map=alpha_map)])

    # Add the `ChannelAlignment` algorithm as the second step of the pipeline.
    if not advanced_parameters.disable_channel_alignment:
        pipeline_steps.append([ChannelAlignment(subset_size, advanced_parameters.inplace_statistics)])

    # Add the `MinMaxQuantization` algorithm as the third step of the pipeline.
    pipeline_steps.append(
        [
            MinMaxQuantization(
                mode=mode,
                preset=preset,
                target_device=target_device,
                subset_size=subset_size,
                model_type=model_type,
                ignored_scope=ignored_scope,
                overflow_fix=advanced_parameters.overflow_fix,
                quantize_outputs=advanced_parameters.quantize_outputs,
                inplace_statistics=advanced_parameters.inplace_statistics,
                batchwise_statistics=advanced_parameters.batchwise_statistics,
                activations_quantization_params=advanced_parameters.activations_quantization_params,
                weights_quantization_params=advanced_parameters.weights_quantization_params,
                activations_range_estimator_params=advanced_parameters.activations_range_estimator_params,
                weights_range_estimator_params=advanced_parameters.weights_range_estimator_params,
                backend_params=advanced_parameters.backend_params,
            )
        ]
    )

    if not advanced_parameters.disable_bias_correction:
        # Add the `FastBiasCorrection` or `BiasCorrection` as additional algorithm
        # inside the third step of the pipeline. It is added after `MinMaxQuantization`
        # algorithm.
        bias_correction_params = advanced_parameters.bias_correction_params
        if fast_bias_correction:
            threshold = FAST_BIAS_CORRECTION_THRESHOLD
            bias_correction_subset_size = subset_size
            bias_correction_cls = FastBiasCorrection
        else:
            threshold = BIAS_CORRECTION_THRESHOLD
            bias_correction_subset_size = max(int(subset_size * 0.2), 1)
            bias_correction_cls = BiasCorrection

        if bias_correction_params.threshold is not None:
            threshold = bias_correction_params.threshold

        pipeline_steps[-1].append(
            bias_correction_cls(
                bias_correction_subset_size,
                threshold,
                bias_correction_params.apply_for_all_nodes,
                advanced_parameters.inplace_statistics,
                advanced_parameters.backend_params,
            )
        )

    return Pipeline(pipeline_steps)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ctypes
import functools
import json
import os
import os.path as osp
from collections import OrderedDict
from copy import deepcopy
from pathlib import Path
from typing import Callable, Dict, List, Tuple, Union

import numpy as np
import pandas as pd
import torch
import torch.nn.parallel
import torch.optim
import torch.utils.data
import torch.utils.data.distributed
from natsort import natsorted
from sklearn.preprocessing import MinMaxScaler
from torch import nn

from nncf.common.hardware.config import HWConfigType
from nncf.common.initialization.batchnorm_adaptation import BatchnormAdaptationAlgorithm
from nncf.common.logging import nncf_logger
from nncf.common.quantization.quantizer_setup import QuantizationPointId
from nncf.common.quantization.structs import NonWeightQuantizerId
from nncf.common.quantization.structs import QuantizerConfig
from nncf.common.quantization.structs import QuantizerId
from nncf.common.quantization.structs import WeightQuantizerId
from nncf.common.utils.debug import DEBUG_LOG_DIR
from nncf.common.utils.debug import is_debug
from nncf.common.utils.os import safe_open
from nncf.config.extractors import extract_bn_adaptation_init_params
from nncf.torch.initialization import PartialDataLoader
from nncf.torch.quantization.algo import ExperimentalQuantizationController
from nncf.torch.quantization.algo import NNCFNetwork
from nncf.torch.quantization.algo import QuantizationController
from nncf.torch.quantization.layers import BaseQuantizer
from nncf.torch.quantization.precision_constraints import HardwareQuantizationConstraints
from nncf.torch.quantization.precision_init.compression_ratio import CompressionRatioCalculator


def find_qid_by_str(qctrl: QuantizationController, qid_str: str) -> QuantizerId:
    for _qid, _q in qctrl.all_quantizations.items():
        if qid_str == str(_qid):
            return _qid
    return None


class ModelSizeCalculator:
    FLOAT_BITWIDTH = ctypes.sizeof(ctypes.c_float) * 8

    def __init__(self, qmodel: NNCFNetwork, per_quantizer_config_space: Dict[QuantizerId, List[QuantizerConfig]]):
        self._bw_space_map = OrderedDict()
        self._nparam_map = OrderedDict()
        for qid, qconfig_space in per_quantizer_config_space.items():
            if isinstance(qid, WeightQuantizerId):
                self._bw_space_map[qid] = [qconf.num_bits for qconf in qconfig_space]
                m = qmodel.nncf.get_containing_module(qid.target_node_name)
                self._nparam_map[qid] = np.prod(m.weight.size())

        self.min_model_size, self.max_model_size = self._calc_min_max_model_size()

        self.fp_model_size = self.get_uniform_bit_model_size(ModelSizeCalculator.FLOAT_BITWIDTH)

    def __call__(self, per_quantizer_bw):
        return self.get_model_size(per_quantizer_bw)

    def _calc_min_max_model_size(self) -> Tuple[np.int64, np.int64]:
        nparam_array = np.array(list(self._nparam_map.values()))
        min_size = np.sum(nparam_array * np.array(list(map(min, self._bw_space_map.values()))))
        max_size = np.sum(nparam_array * np.array(list(map(max, self._bw_space_map.values()))))
        return min_size, max_size

    def get_uniform_bit_model_size(self, uniform_bitwidth: int) -> np.int64:
        return np.sum(np.array(list(self._nparam_map.values()) * uniform_bitwidth))

    def get_model_size(self, per_quantizer_bw: Dict[QuantizerId, int]) -> np.int64:
        model_size = 0
        for qid, nparam in self._nparam_map.items():
            if qid in per_quantizer_bw:
                model_size += nparam * per_quantizer_bw[qid]
            else:
                nncf_logger.warning(
                    f"[ModelSizeCalculator] Missing Bitwidth of QID: {str(qid)}, "
                    f"using {ModelSizeCalculator.FLOAT_BITWIDTH} bits"
                )
                model_size += nparam * ModelSizeCalculator.FLOAT_BITWIDTH
        return model_size

    def get_model_size_ratio(self, per_quantizer_bw: Dict[QuantizerId, int]) -> np.float64:
        return self.get_model_size(per_quantizer_bw) / self.fp_model_size


class QuantizationEnvParams:
    def __init__(
        self,
        compression_ratio: float,
        eval_subset_ratio: float,
        skip_constraint: bool,
        performant_bw: bool,
        finetune: bool,
        bits: List[int],
        dump_init_precision_data: bool = False,
        log_dir: str = None,
    ):
        self.compression_ratio = compression_ratio
        self.eval_subset_ratio = eval_subset_ratio
        self.skip_constraint = skip_constraint
        self.performant_bw = performant_bw
        self.finetune = finetune
        self.bits = bits
        self.dump_init_precision_data = dump_init_precision_data
        self.log_dir = log_dir


class QuantizationEnv:
    def __init__(
        self,
        model: NNCFNetwork,
        quantization_controller: ExperimentalQuantizationController,
        hw_precision_constraints: HardwareQuantizationConstraints,
        eval_loader: torch.utils.data.DataLoader,
        eval_fn: Callable[[nn.Module, torch.utils.data.DataLoader], float],
        hw_config_type: HWConfigType,
        params: QuantizationEnvParams,
    ):
        nncf_logger.info("[Q.Env] Instantiating NNCF Quantization Environment...")
        self.qctrl = quantization_controller
        self.qmodel = model
        self.eval_loader = eval_loader
        self.eval_fn = eval_fn
        self._hw_precision_constraints = hw_precision_constraints
        self._bn_adaptation = None

        self.model_name = self.qmodel.__class__.__name__

        # Check and only proceed if target device is supported by Q.Env
        self.hw_cfg_type = hw_config_type
        assert self.hw_cfg_type in [None, HWConfigType.NPU]

        # Set target compression ratio
        self.compression_ratio = params.compression_ratio

        self.eval_loader = PartialDataLoader(self.eval_loader, iter_ratio=params.eval_subset_ratio)

        # Bool to disable hard resource constraint
        self.skip_constraint = params.skip_constraint

        # Bool to enable bw alignment of adj. Q group to lower precision
        self.performant_bw = params.performant_bw

        # Bool to enable fine-tuning in each episode. Placeholder for now
        self.finetune = False

        # Counter for number of evaluate_strategy calls
        self._n_eval = 0

        # Configure search space for precision according to target device
        if self.hw_cfg_type is None:
            self.model_bitwidth_space = params.bits
        elif self.hw_cfg_type is HWConfigType.NPU:
            self.model_bitwidth_space = self._hw_precision_constraints.get_all_unique_bitwidths()
        self.model_bitwidth_space = sorted(list(self.model_bitwidth_space))

        # Create mapping of QuantizerId to the space of the corresponding quantizer's allowed qconfigs
        self.qconfig_space_map: Dict[QuantizerId, List[QuantizerConfig]] = OrderedDict.fromkeys(
            self.qctrl.all_quantizations.keys()
        )
        if self.hw_cfg_type is None:
            for qid in self.qconfig_space_map:
                conf = self.qctrl.all_quantizations[qid].get_quantizer_config()
                conf_list_to_set = []
                for bit in self.model_bitwidth_space:
                    bit_adjusted_conf = deepcopy(conf)
                    bit_adjusted_conf.num_bits = bit
                    conf_list_to_set.append(bit_adjusted_conf)
                self.qconfig_space_map[qid] = conf_list_to_set
        else:
            for qid in self.qconfig_space_map:
                conf_list_to_set = []
                bw_vs_qconfigs_dict = self._hw_precision_constraints.get_bitwidth_vs_qconfigs_dict(qid)
                for bitwidth, qconf_list in bw_vs_qconfigs_dict.items():
                    target_qconf = qconf_list[0]
                    if len(qconf_list) > 1:
                        nncf_logger.warning(
                            f"Received multiple quantizer configurations "
                            f"{';'.join([str(qconf) for qconf in qconf_list])} for same bitwidth {bitwidth} for "
                            f"quantizer {str(qid)} - AutoQ can currently only choose among bitwidths, but not within "
                            f"quantizer configuration space with the same bitwidths. Selecting {str(target_qconf)} as "
                            f"the target configuration for bitwidth {bitwidth}"
                        )
                    conf_list_to_set.append(target_qconf)

                self.qconfig_space_map[qid] = conf_list_to_set

        # Quantizer Master Table Creation
        self.groups_of_adjacent_quantizers = self.qctrl._groups_of_adjacent_quantizers
        self.quantizer_table = self._create_quantizer_table()

        # Create master dataframe to keep track of quantizable layers and their attributes
        self.master_df, self.state_list = self._get_state_space(self.qctrl, self.qmodel, self.quantizer_table)
        if self.master_df.isnull().values.any():
            raise ValueError("Q.Env Master Dataframe has null value(s)")

        assert len(self.quantizer_table) == len(
            self.qctrl.all_quantizations
        ), "Number of Quantizer is not tally between quantizer table and quantization controller"

        # MinMaxScaler for State Embedding
        self.state_scaler = MinMaxScaler()
        self.state_scaler.fit(self.master_df[self.state_list])

        # Mapping required for quantizer BW alignment flow
        self.adjq_groupwise_intersecting_bw_space = self._create_map_of_adjq_groupid_to_common_bw_space()
        self.adjq_groupwise_df_lut_keys = self._create_map_of_adjq_groupid_to_df_lut_keys()

        # Model Size Calculation
        self.model_size_calculator = ModelSizeCalculator(self.qmodel, self.qconfig_space_map)
        self.orig_model_size = self.model_size_calculator.fp_model_size
        self.min_model_size = self.model_size_calculator.min_model_size
        self.max_model_size = self.model_size_calculator.max_model_size
        self.target_model_size = self.orig_model_size * self.compression_ratio

        if self.target_model_size < self.min_model_size and self.target_model_size > self.max_model_size:
            raise ValueError(
                "Model Size Ratio {} is out of bound ({}, {})".format(
                    self.compression_ratio,
                    self.min_model_size / self.orig_model_size,
                    self.max_model_size / self.orig_model_size,
                )
            )

        # Compression Ratio Calculation (BOP relative to 8-bit)
        self.compression_ratio_calculator = CompressionRatioCalculator(
            self.qmodel.nncf.get_flops_per_module(),
            self.qctrl.get_quantizer_setup_for_current_state(),
            self.qctrl.groups_of_adjacent_quantizers.weight_qp_id_per_activation_qp_id,
        )

        # Evaluate and store metric score of pretrained model
        self._evaluate_pretrained_model()
        self.qmodel_init_sd = deepcopy(self.qmodel.state_dict())

        self.reset()

        self._dump_autoq_data = params.dump_init_precision_data
        if self._dump_autoq_data or is_debug():
            dump_dir = params.log_dir
            if dump_dir is None:
                dump_dir = DEBUG_LOG_DIR
            self.dump_dir = Path(dump_dir) / Path("autoq_env_dump")
            self.dump_dir.mkdir(parents=True, exist_ok=True)
            # Serialize Q.Env information. Note that these functions should be at the end of Q.Env Initialization.
            self._dump_master_df()
            self._dump_quantized_graph()
            self._dump_groups_of_adjacent_quantizers()

        # End of QuantizationEnv.__init__()
        # --------------------------------------------------------------------------------------------------------------

    def reset(self):
        self.collected_strategy = []
        self.master_df["action"] = max(self.model_bitwidth_space)
        self.master_df["prev_action"] = 0
        self.master_df["unconstrained_action"] = 0

    def _create_quantizer_table(self) -> pd.DataFrame:
        def get_hook(qid, exec_order_list):
            def register_quantizer_exec_order(module, input_, output, qid, exec_order_list):
                exec_order_list.append(qid)

            return functools.partial(register_quantizer_exec_order, qid=qid, exec_order_list=exec_order_list)

        # Create a mapping of qid to its adjacent quantizer group id
        adjq_gid_map = OrderedDict.fromkeys(self.qctrl.all_quantizations.keys())
        for qid in self.qctrl.all_quantizations:
            adjq_gid_map[qid] = self.groups_of_adjacent_quantizers.get_group_id_for_quantizer(qid)

        assert (
            len(set(self.qconfig_space_map.keys()) - set(adjq_gid_map.keys())) == 0
        ), "both qconfig_space_map and adjq_gid_map must have exact keys."

        # By design, AutoQ requires quantizers in execution order.
        # RL assumes that state satisfies Markov assumption in which
        # the future is independent of the past given current state.
        # Stated differently, curret state should represent well of historical dynamics.
        # Given sequential nature of NN, state transition in the order of
        # quantizer being executed is a natural design to conform the assumption.
        quantizers_in_exec_order = []
        hooklist = []
        for qid, qmod in self.qctrl.all_quantizations.items():
            hooklist.append(qmod.register_forward_hook(get_hook(qid, quantizers_in_exec_order)))
        self.qmodel.nncf.do_dummy_forward(force_eval=True)
        for h in hooklist:
            h.remove()

        d = OrderedDict()
        for qid in quantizers_in_exec_order:
            idx_str = str(qid)
            gid = adjq_gid_map[qid]

            d[idx_str] = OrderedDict()
            d[idx_str]["qid"] = str(qid)
            d[idx_str]["gid"] = gid
            d[idx_str]["qconf_space"] = self.qconfig_space_map[qid]
            d[idx_str]["qp_id_set"] = self.qctrl.module_id_to_qp_id_translation_dict[qid]
            d[idx_str]["state_scope"] = qid.target_node_name

        # quantizer_table index is QuantizerId in string prepended with its quantize node id in NNCFGraph
        df = pd.DataFrame.from_dict(d, orient="index")
        df["qid_obj"] = df["qid"].apply(lambda x: find_qid_by_str(self.qctrl, x))
        df["qmodule"] = df["qid_obj"].apply(lambda x: self.qctrl.all_quantizations[x])
        df["is_wt_quantizer"] = df["qid_obj"].apply(lambda x: x in self.qctrl.weight_quantizers)
        df["state_module"] = df["state_scope"].apply(self.qmodel.nncf.get_containing_module)

        return df

    def _get_state_space(
        self,
        quantization_controller: QuantizationController,
        quantized_model: NNCFNetwork,
        quantizer_table: pd.DataFrame,
    ) -> Tuple[pd.DataFrame, List[str]]:
        def annotate_learnable_module_io_shape(model):
            def annotate_io_shape(module, input_, output):
                if hasattr(module, "weight") or isinstance(module, BaseQuantizer):
                    module.input_shape_ = input_[0].shape
                    module.output_shape_ = output.shape

            hook_list = [m.register_forward_hook(annotate_io_shape) for n, m in model.named_modules()]
            model.nncf.do_dummy_forward(force_eval=True)
            for h in hook_list:
                h.remove()

        annotate_learnable_module_io_shape(quantized_model)

        # State Embedding Extraction
        # ---------------------------
        df = quantizer_table
        layer_attr_df = df.apply(self._get_layer_attr, axis=1)
        layer_attr_df["layer_idx"] = np.array(range(len(layer_attr_df)))
        layer_attr_df["weight_quantizer"] = df["is_wt_quantizer"].astype("float")
        state_list = layer_attr_df.columns.to_list()

        # create master dataframe
        master_df = pd.concat([df, layer_attr_df], axis="columns")

        # Annotate a min and a max value in prev_action before minmaxscaler fitting
        master_df.loc[master_df.index[0], "prev_action"] = max(self.model_bitwidth_space)
        master_df.loc[master_df.index[-1], "prev_action"] = min(self.model_bitwidth_space)

        # add GEMM Ops to weight quantizer
        master_df["n_op"] = master_df["state_scope"].map(self.qmodel.nncf.get_flops_per_module())
        master_df["n_op"] = master_df["n_op"].fillna(0)

        return master_df, state_list

    def _get_layer_attr(self, row: pd.Series) -> pd.Series:
        m = row.state_module
        qid = row.qid_obj
        feature = OrderedDict()

        if isinstance(qid, WeightQuantizerId):
            if isinstance(m, (nn.Conv2d, nn.ConvTranspose2d)):
                feature["conv_dw"] = int(m.weight.shape[1] == m.groups)  # 1.0 for depthwise, 0.0 for other conv2d
                feature["cin"] = m.weight.shape[1]
                feature["cout"] = m.weight.shape[0]
                feature["stride"] = m.stride[0]
                feature["kernel"] = m.kernel_size[0]
                feature["param"] = np.prod(m.weight.size())
                feature["ifm_size"] = np.prod(m.input_shape_[-2:])  # H*W
                feature["prev_action"] = 0.0  # placeholder

            elif isinstance(m, nn.Linear):
                feature["conv_dw"] = 0.0
                feature["cin"] = m.in_features
                feature["cout"] = m.out_features
                feature["stride"] = 0.0
                feature["kernel"] = 1.0
                feature["param"] = np.prod(m.weight.size())
                feature["ifm_size"] = np.prod(m.input_shape_[-1])  # feature elements
                feature["prev_action"] = 0.0  # placeholder

            else:
                raise NotImplementedError("State embedding extraction of {}".format(m.__class__.__name__))

        elif isinstance(qid, NonWeightQuantizerId):
            qmod = self.qctrl.all_quantizations[qid]
            input_shape = qmod.input_shape_
            output_shape = qmod.output_shape_
            feature["cin"] = input_shape[1] if len(input_shape) == 4 else input_shape[-1]
            feature["cout"] = output_shape[1] if len(output_shape) == 4 else output_shape[-1]
            feature["ifm_size"] = np.prod(input_shape[-2:]) if len(input_shape) == 4 else input_shape[-1]
            feature["conv_dw"] = 0.0
            feature["stride"] = 0.0
            feature["kernel"] = 0.0
            feature["param"] = 0.0
            feature["prev_action"] = 0.0

            if len(input_shape) != 4 and len(input_shape) != 2:
                raise NotImplementedError("A design is required to cater this scenario. Pls. report to maintainer")
        else:
            raise ValueError("qid is an instance of unexpected class {}".format(qid.__class__.__name__))

        return pd.Series(feature)

    def _create_map_of_adjq_groupid_to_common_bw_space(self) -> Dict:
        # Extracting common bitwidth space per group of quantizer
        bwassigner_df = deepcopy(self.master_df)
        bwassigner_df["bw_space"] = list(map(lambda x: [qc.num_bits for qc in x], bwassigner_df.qconf_space.values))

        adjq_groupwise_intersecting_bw_space = {}
        for i, _ in enumerate(self.groups_of_adjacent_quantizers):
            list_of_bw_space = []

            for aq in self.groups_of_adjacent_quantizers[i].activation_quantizers:
                bw_space = bwassigner_df.bw_space[bwassigner_df.qid == str(aq[0])][0]
                list_of_bw_space.append(bw_space)

            for wq in self.groups_of_adjacent_quantizers[i].weight_quantizers:
                bw_space = bwassigner_df.bw_space[bwassigner_df.qid == str(wq[0])][0]
                list_of_bw_space.append(bw_space)

            intersecting_bw_space = set.intersection(*map(set, list_of_bw_space))
            adjq_groupwise_intersecting_bw_space[i] = intersecting_bw_space

        return adjq_groupwise_intersecting_bw_space

    def _create_map_of_adjq_groupid_to_df_lut_keys(self) -> Dict:
        adjq_groupwise_df_lut_keys = {}

        for i, _ in enumerate(self.groups_of_adjacent_quantizers):
            group_members = []
            for _, aq in enumerate(self.groups_of_adjacent_quantizers[i].activation_quantizers):
                group_members.append(self.master_df.index[self.master_df.qid == str(aq[0])][0])
            for _, wq in enumerate(self.groups_of_adjacent_quantizers[i].weight_quantizers):
                group_members.append(self.master_df.index[self.master_df.qid == str(wq[0])][0])
            adjq_groupwise_df_lut_keys[i] = natsorted(group_members)

        return adjq_groupwise_df_lut_keys

    def _evaluate_pretrained_model(self):
        nncf_logger.info("[Q.Env] Evaluating Pretrained Model")
        self.qctrl.disable_weight_quantization()
        self.qctrl.disable_activation_quantization()

        with torch.no_grad():
            self.pretrained_score = self.eval_fn(self.qmodel, self.eval_loader)
            nncf_logger.info(f"Pretrained Score: {self.pretrained_score:.3f}")

        self.qctrl.enable_weight_quantization()
        self.qctrl.enable_activation_quantization()
        self.qmodel.nncf.rebuild_graph()

    def _run_batchnorm_adaptation(self):
        if self._bn_adaptation is None:
            self._bn_adaptation = BatchnormAdaptationAlgorithm(
                **extract_bn_adaptation_init_params(self.qctrl.config, "quantization")
            )
        self._bn_adaptation.run(self.qctrl.model)

    def _run_quantization_pipeline(self, finetune=False) -> float:
        if self.qctrl.config:
            self._run_batchnorm_adaptation()

        if finetune:
            raise NotImplementedError("Post-Quantization fine tuning is not implemented.")
        with torch.no_grad():
            quantized_score = self.eval_fn(self.qmodel, self.eval_loader)
            nncf_logger.info(f"[Q.Env] Quantized Score: {quantized_score:.3f}")
        return quantized_score

    def _get_quantizer_bitwidth(self) -> Dict[BaseQuantizer, int]:
        assert (
            len(set(self.model_bitwidth_space) - set(self.master_df.action.values)) >= 0
        ), "there is bitwidth choice not within model bitwidth space"
        return OrderedDict(zip(self.master_df.qid_obj, self.master_df.action))

    def _constrain_model_size(self, collected_strategy: List, skip=False) -> List:
        def lower_bitwidth(bw: int, qconf_space: List[QuantizerConfig]) -> int:
            bw_space = [qconf.num_bits for qconf in qconf_space]
            assert bw in bw_space
            sorted_bw_space = sorted(bw_space)
            return sorted_bw_space[sorted_bw_space.index(bw) - 1] if sorted_bw_space.index(bw) > 0 else bw

        # This function acts on self.master_df['action']
        self.master_df["action"] = collected_strategy

        if skip is not True:
            self.master_df["unconstrained_action"] = self.master_df["action"]

            current_model_size = self.model_size_calculator(self._get_quantizer_bitwidth())

            while self.min_model_size < current_model_size and self.target_model_size < current_model_size:
                for _, nodestr in enumerate(reversed(self.master_df.index.tolist())):
                    if self.master_df.loc[nodestr, "is_wt_quantizer"]:
                        bw_choice, qconf_space = self.master_df.loc[nodestr, ["action", "qconf_space"]]
                        new_bw = lower_bitwidth(bw_choice, qconf_space)
                        self.master_df.loc[nodestr, "action"] = new_bw if new_bw != bw_choice else bw_choice

                    current_model_size = self.model_size_calculator(self._get_quantizer_bitwidth())
                    if current_model_size <= self.target_model_size:
                        break
        else:
            nncf_logger.info("[Q.Env] Skipping the model size constraint.")

        return self.master_df["action"].tolist()

    def reward(self, acc: float, model_ratio: float) -> float:
        def order_of_magnitude(number):
            return np.floor(np.math.log(abs(number), 10))

        if self.pretrained_score == 0:
            return acc
        order = order_of_magnitude(self.pretrained_score)
        return acc * (10 ** (-order))

    def step(self, action: Union[int, float]) -> Tuple:
        currently_processed_qconf_idx = len(self.collected_strategy)

        def is_final_step():
            return len(self.collected_strategy) == len(self.master_df)

        # Ensure action is in the quantizer's bitwidth space
        current_qconf_space = self.master_df.qconf_space[currently_processed_qconf_idx]
        current_bw_space = [qconf.num_bits for qconf in current_qconf_space]
        if action not in current_bw_space:
            closest_bw_idx = np.argmin(np.abs(action - np.array(current_bw_space)))
            action = current_bw_space[closest_bw_idx]

        self.collected_strategy.append(action)

        if not is_final_step():
            info_set = {}
            reward = 0
            self.set_next_step_prev_action(len(self.collected_strategy), action)
            obs = self.get_normalized_obs(len(self.collected_strategy))
            done = False
            return obs, reward, done, info_set

        return self.evaluate_strategy(self.collected_strategy, skip_constraint=self.skip_constraint)

    def select_config_for_actions(self, actions) -> Dict[QuantizationPointId, QuantizerConfig]:
        retval: Dict[QuantizationPointId, QuantizerConfig] = OrderedDict()
        for action, qp_id_set, qconf_space in zip(actions, self.master_df["qp_id_set"], self.master_df["qconf_space"]):
            matches = []
            for qconf in qconf_space:
                if qconf.num_bits == action:
                    matches.append(qconf)
            assert len(matches) == 1
            for qp_id in qp_id_set:
                retval[qp_id] = matches[0]
        return retval

    def evaluate_strategy(self, collected_strategy: List, skip_constraint=True) -> Tuple:
        assert len(collected_strategy) == len(self.master_df)
        if skip_constraint is not True:
            collected_strategy = self._constrain_model_size(collected_strategy)
        self.master_df["action"] = collected_strategy  # This must be after constraint

        if self.performant_bw:
            self._align_bw_action()
            configs_to_set = self.select_config_for_actions(self.master_df["action_aligned"])

            if self._dump_autoq_data or is_debug():
                self._dump_adjacent_quantizer_group_alignment()

            self.master_df["action"] = self.master_df["action_aligned"]
        else:
            configs_to_set = self.select_config_for_actions(self.master_df["action"])

        self._apply_quantizer_configs_to_model(configs_to_set)

        for idx, qid in zip(self.master_df.index, self.master_df["qid"]):
            nncf_logger.info(
                f"[Q.Env] {str(self.qctrl.all_quantizations[find_qid_by_str(self.qctrl, qid)]):50} | {idx}"
            )

        quantized_score = self._run_quantization_pipeline(finetune=self.finetune)

        current_model_size = self.model_size_calculator(self._get_quantizer_bitwidth())
        current_model_ratio = self.model_size_calculator.get_model_size_ratio(self._get_quantizer_bitwidth())

        current_model_bop_ratio = self.compression_ratio_calculator.run_for_quantizer_setup(
            self.qctrl.get_quantizer_setup_for_current_state()
        )

        reward = self.reward(quantized_score, current_model_ratio)

        info_set = {
            "model_ratio": current_model_ratio,
            "accuracy": quantized_score,
            "model_size": current_model_size,
            "bop_ratio": current_model_bop_ratio,
        }

        obs = self.get_normalized_obs(len(collected_strategy) - 1)
        done = True
        self._n_eval += 1

        return obs, reward, done, info_set

    def set_next_step_prev_action(self, idx, action):
        self.master_df.loc[self.master_df.index[idx], "prev_action"] = action

    def get_normalized_obs(self, idx: int) -> pd.Series:
        _df = self.master_df.loc[self.master_df.index, self.state_list]
        _df.loc[_df.index, self.state_list] = self.state_scaler.transform(_df[self.state_list])
        return _df.iloc[idx]

    def _apply_quantizer_configs_to_model(self, qid_vs_qconfig_map: Dict[QuantizationPointId, QuantizerConfig]):
        new_quantizer_setup = self.qctrl.get_quantizer_setup_for_current_state()
        for qp_id, qconf in qid_vs_qconfig_map.items():
            new_quantizer_setup.quantization_points[qp_id].qconfig = qconf
        self.qmodel.load_state_dict(self.qmodel_init_sd)
        had_to_regenerate = self.qctrl.is_new_setup_requires_regeneration(new_quantizer_setup)
        self.qctrl, self.qmodel = self.qctrl.apply_new_quantizer_setup(new_quantizer_setup)
        if had_to_regenerate:
            self.qmodel_init_sd = deepcopy(self.qmodel.state_dict())

        # The QuantizerId's may have changed after the new quantizer setup application, but
        # QuantizationPointId's should not have. Will use this to update the qids in the master table.
        for qid, qp_id_set in self.qctrl.module_id_to_qp_id_translation_dict.items():
            self.master_df.loc[self.master_df.qp_id_set == qp_id_set].qid = str(qid)

    def _dump_master_df(self):
        self.master_df.drop("state_module", axis=1).to_csv(
            osp.join(self.dump_dir, self.model_name + "_quantizable_state_table.csv"), index_label="nodestr"
        )

    def _dump_quantized_graph(self):
        self.qmodel.nncf.get_graph().visualize_graph(osp.join(self.dump_dir, "qenv_graph.dot"))

    def _dump_groups_of_adjacent_quantizers(self):
        adj_quantizer_groups = []

        for i, _ in enumerate(self.groups_of_adjacent_quantizers):
            group_members = []
            for _, aq in enumerate(self.groups_of_adjacent_quantizers[i].activation_quantizers):
                group_members.append(self.master_df.index[self.master_df.qid == str(aq[0])][0])
            for _, wq in enumerate(self.groups_of_adjacent_quantizers[i].weight_quantizers):
                group_members.append(self.master_df.index[self.master_df.qid == str(wq[0])][0])
            adj_quantizer_groups.append(natsorted(group_members))

        with safe_open(self.dump_dir / "{}_groups_of_adjacent_quantizers.json".format(self.model_name), "w") as DUMP_FH:
            json.dump(natsorted(adj_quantizer_groups), DUMP_FH, indent=4)

    def _align_bw_action(self):
        # align bw action per group of adjacent quantizer
        # this alignment aims to realize GEMM compute in a lower precision

        self.master_df["action_aligned"] = 0

        for i, _ in enumerate(self.groups_of_adjacent_quantizers):
            # Collect all actions of a group
            list_of_action = []

            for _, aq in enumerate(self.groups_of_adjacent_quantizers[i].activation_quantizers):
                list_of_action.append(self.master_df.action[self.master_df.qid == str(aq[0])][0])

            for _, wq in enumerate(self.groups_of_adjacent_quantizers[i].weight_quantizers):
                list_of_action.append(self.master_df.action[self.master_df.qid == str(wq[0])][0])

            # Get the minimum prediction bw of a group
            group_min_predicted_bw = min(list_of_action)

            # Access and get the intersecting bw among all quantizers in a group
            intersecting_bw_space = self.adjq_groupwise_intersecting_bw_space[i]

            # Determine the lowest realizable hardware precision given current action of the groups
            if group_min_predicted_bw not in intersecting_bw_space:
                group_final_min_bw = min(intersecting_bw_space)
            else:
                group_final_min_bw = group_min_predicted_bw

            # Assignment Routine
            for _, aq in enumerate(self.groups_of_adjacent_quantizers[i].activation_quantizers):
                self.master_df.loc[self.master_df.qid == str(aq[0]), "action_aligned"] = group_final_min_bw

            for _, wq in enumerate(self.groups_of_adjacent_quantizers[i].weight_quantizers):
                if self.master_df.loc[self.master_df.qid == str(wq[0]), "action"][0] > group_final_min_bw:
                    self.master_df.loc[self.master_df.qid == str(wq[0]), "action_aligned"] = group_final_min_bw
                else:
                    self.master_df.loc[self.master_df.qid == str(wq[0]), "action_aligned"] = self.master_df.loc[
                        self.master_df.qid == str(wq[0]), "action"
                    ][0]

    def _dump_adjacent_quantizer_group_alignment(self):
        list_of_dump_dict = []
        for i, _ in enumerate(self.groups_of_adjacent_quantizers):
            list_of_dump_dict.append(
                self.master_df.loc[self.adjq_groupwise_df_lut_keys[i], ["action", "action_aligned"]].to_dict()
            )

        os.makedirs(self.dump_dir / "bw_alignment", exist_ok=True)
        with safe_open(self.dump_dir / "bw_alignment/{0:03d}_bw_alignment.json".format(self._n_eval), "w") as DUMP_FH:
            json.dump(list_of_dump_dict, DUMP_FH, indent=4)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List

import numpy as np
import tensorflow as tf

import nncf
from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNodeName
from nncf.common.logging import nncf_logger
from nncf.tensorflow.graph.metatypes.common import GENERAL_CONV_LAYER_METATYPES
from nncf.tensorflow.graph.metatypes.common import LINEAR_LAYER_METATYPES
from nncf.tensorflow.graph.metatypes.keras_layers import TFBatchNormalizationLayerMetatype
from nncf.tensorflow.graph.metatypes.matcher import get_keras_layer_metatype
from nncf.tensorflow.graph.utils import get_original_name_and_instance_idx
from nncf.tensorflow.graph.utils import unwrap_layer
from nncf.tensorflow.layers.data_layout import get_input_channel_axis
from nncf.tensorflow.layers.data_layout import get_weight_channel_axis
from nncf.tensorflow.layers.wrapper import NNCFWrapper


def get_filter_axis(layer: NNCFWrapper, weight_attr: str) -> int:
    channel_axes = get_weight_channel_axis(layer, weight_attr)
    filter_axis = channel_axes[0] if isinstance(channel_axes, tuple) else channel_axes
    return filter_axis


def get_filters_num(layer: NNCFWrapper):
    layer_metatype = get_keras_layer_metatype(layer)
    if len(layer_metatype.weight_definitions) != 1:
        raise ValueError(f"Could not calculate the number of filters for the layer {layer.layer.name}.")

    weight_def = layer_metatype.weight_definitions[0]
    weight_attr = weight_def.weight_attr_name

    if layer_metatype is TFBatchNormalizationLayerMetatype and not layer.layer.scale:
        nncf_logger.debug(
            "Fused gamma parameter encountered in BatchNormalization layer. "
            "Using beta parameter instead to calculate the number of filters."
        )
        weight_attr = "beta"

    filter_axis = get_filter_axis(layer, weight_attr)
    filters_num = layer.layer_weights[weight_attr].shape[filter_axis]
    return filters_num


def is_valid_shape(shape):
    if shape is None:
        return False
    if None in shape:
        return False
    return True


def broadcast_filter_mask(filter_mask, shape, dim):
    broadcasted_filter_mask = tf.zeros(shape)
    meta_shape = np.ones(len(shape), dtype=np.int64)
    meta_shape[dim] = filter_mask.shape[0]
    broadcasted_filter_mask += tf.reshape(filter_mask, tuple(meta_shape))
    return broadcasted_filter_mask


def collect_output_shapes(model: "NNCFNetwork", graph: NNCFGraph) -> Dict[NNCFNodeName, List[int]]:  # noqa: F821
    """
    Collects output dimension shapes for convolutions and fully connected layers
    from the connected edges in the NNCFGraph.

    :param graph: NNCFGraph.
    :return: Dictionary of output dimension shapes. E.g {node_name: (height, width)}
    """
    layers_out_shapes = {}
    for node in graph.get_nodes_by_metatypes(GENERAL_CONV_LAYER_METATYPES):
        node_name, node_index = get_original_name_and_instance_idx(node.node_name)
        layer = model.get_layer(node_name)
        layer_ = unwrap_layer(layer)

        channel_axis = get_input_channel_axis(layer)
        dims_slice = (
            slice(channel_axis - layer_.rank, channel_axis)
            if layer.data_format == "channels_last"
            else slice(channel_axis + 1, None)
        )
        in_shape = layer.get_input_shape_at(node_index)[dims_slice]
        out_shape = layer.get_output_shape_at(node_index)[dims_slice]

        if not is_valid_shape(in_shape) or not is_valid_shape(out_shape):
            raise nncf.ValidationError(f"Input/output shape is not defined for layer `{layer.name}` ")

        layers_out_shapes[node.node_name] = out_shape

    for node in graph.get_nodes_by_metatypes(LINEAR_LAYER_METATYPES):
        node_name, node_index = get_original_name_and_instance_idx(node.node_name)
        layer = model.get_layer(node_name)

        in_shape = layer.get_input_shape_at(node_index)[1:]
        out_shape = layer.get_output_shape_at(node_index)[1:]

        if not is_valid_shape(in_shape) or not is_valid_shape(out_shape):
            raise nncf.ValidationError(f"Input/output shape is not defined for layer `{layer.name}` ")

        layers_out_shapes[node.node_name] = out_shape
    return layers_out_shapes



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from typing import List, Optional, Set, Union

import nncf
from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.graph import NNCFNodeName
from nncf.common.logging import nncf_logger
from nncf.common.quantization.structs import QuantizerId
from nncf.scopes import IgnoredScope
from nncf.scopes import convert_ignored_scope_to_list


def matches_any(tested_str: str, strs_to_match_to: Union[List[str], Set[str], str]) -> bool:
    """
    Return True if tested_str matches at least one element in strs_to_match_to.

    :param tested_str: One of the supported entity types to be matched - currently possible to pass either
        NNCFNodeName (to refer to the original model operations) or QuantizerId (to refer to specific quantizers).
    :param strs_to_match_to: A list or set of strings specifying for the serializable_id. Entries of the strings
        may be prefixed with `{re}` to enable regex matching.

    :return: A boolean value specifying whether a tested_str should matches at least one element
        in strs_to_match_to.
    """

    if strs_to_match_to is None:
        return False

    str_list = [strs_to_match_to] if isinstance(strs_to_match_to, str) else strs_to_match_to
    for item in str_list:
        if "{re}" in item:
            regex = item.replace("{re}", "")
            if re.search(regex, tested_str):
                return True
        else:
            if tested_str == item:
                return True
    return False


def should_consider_scope(
    serializable_id: Union[QuantizerId, NNCFNodeName],
    ignored_scopes: Union[List[str], Set[str]],
    target_scopes: Optional[List[str]] = None,
) -> bool:
    """
    Used when an entity arising during compression has to be compared to an allowlist or a denylist of strings.

    :param serializable_id: One of the supported entity types to be matched - currently possible to pass either
        NNCFNodeName (to refer to the original model operations) or QuantizerId (to refer to specific quantizers)
    :param ignored_scopes: A list or set of strings specifying a denylist for the serializable_id. Entries of the list
        may be prefixed with `{re}` to enable regex matching.
    :param target_scopes: A list of strings specifying an allowlist for the serializable_id. Entries of the list
        may be prefixed with `{re}` to enable regex matching.

    :return: A boolean value specifying whether a serializable_id should be considered (i.e. "not ignored", "targeted")
    """
    string_id = str(serializable_id)
    return (target_scopes is None or matches_any(string_id, target_scopes)) and not matches_any(
        string_id, ignored_scopes
    )


def get_not_matched_scopes(scope: Union[List[str], str, IgnoredScope], nodes: List[NNCFNode]) -> List[str]:
    """
    Return list of scope that do not match node list.

    :param scope: List of ignored/target scope or instance of IgnoredScope.
    :param graph: The model graph.

    :return : List of not matched scopes.
    """

    if isinstance(scope, str):
        patterns = [scope]
    elif isinstance(scope, IgnoredScope):
        patterns = convert_ignored_scope_to_list(scope)
    else:
        patterns = scope

    if not patterns:
        return []

    matched_patterns = set()
    for node in nodes:
        for pattern in patterns:
            if matches_any(node.node_name, pattern):
                matched_patterns.add(pattern)
    return list(set(patterns) - matched_patterns)


def check_scopes_in_graph(
    graph: NNCFGraph,
    ignored_scopes: Union[IgnoredScope, List[str]],
    target_scopes: Optional[List[str]] = None,
    validate_scopes: bool = True,
) -> None:
    """
    Raise RuntimeError in case if ignored/target scope names do not match model graph.

    :param graph: The model graph.
    :param ignored_scopes: The instance of IgnoredScope or a list of strings specifying a denylist
        for the serializable_id.
    :param target_scopes: A list of strings specifying an allowlist for the serializable_id.
    :param validate_scopes: If set to True, then a RuntimeError will be raised if the names of the
      ignored/target scopes do not match the names of the scopes in the model graph.
    """
    node_list = graph.get_all_nodes()
    not_matched_ignored_scopes = get_not_matched_scopes(ignored_scopes, node_list)
    not_matched_target_scopes = get_not_matched_scopes(target_scopes, node_list)

    if not_matched_ignored_scopes or not_matched_target_scopes:
        err_message = (
            "No match has been found among the model operations "
            "for the following ignored/target scope definitions:\n"
        )
        if not_matched_ignored_scopes:
            err_message += f" - ignored_scope: {not_matched_ignored_scopes}\n"
        if not_matched_target_scopes:
            err_message += f" - target_scope: {not_matched_target_scopes}\n"

        err_message += (
            "Refer to the original_graph.dot to discover the operations "
            "in the model currently visible to NNCF and specify the ignored/target "
            "scopes in terms of the names there."
        )

        if validate_scopes:
            raise nncf.ValidationError(err_message)
        nncf_logger.info(err_message)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from nncf.quantization.advanced_parameters import AdvancedQuantizationParameters


class BackendParameters:
    COMPRESS_WEIGHTS = "compress_weights"
    STAT_REQUESTS_NUMBER = "stat_requests_number"
    EVAL_REQUESTS_NUMBER = "eval_requests_number"
    ACTIVATIONS = "activations"
    WEIGHTS = "weights"
    LEVEL_LOW = "level_low"
    LEVEL_HIGH = "level_high"


def is_weight_compression_needed(advanced_parameters: Optional[AdvancedQuantizationParameters]) -> bool:
    """
    Determines whether weight compression is needed based on the provided
    advanced quantization parameters.

    :param advanced_parameters: Advanced quantization parameters.
    :return: True if weight compression is needed, False otherwise.
    """
    if advanced_parameters is not None and advanced_parameters.backend_params is not None:
        return advanced_parameters.backend_params.get(BackendParameters.COMPRESS_WEIGHTS, True)
    return True



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf

from nncf.tensorflow.functions import logit
from nncf.tensorflow.functions import st_threshold


def binary_mask(mask):
    return tf.round(tf.math.sigmoid(mask))


def st_binary_mask(mask):
    return st_threshold(tf.math.sigmoid(mask))


def calc_rb_binary_mask(mask, seed, eps=0.01):
    uniform = tf.random.stateless_uniform(mask.shape, seed=seed, minval=0, maxval=1)
    mask = mask + logit(tf.clip_by_value(uniform, eps, 1 - eps))
    return st_binary_mask(mask)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect
import sys
from typing import List, Tuple

import tensorflow as tf

import nncf
from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.graph import NNCFNodeName
from nncf.tensorflow.graph.metatypes.keras_layers import TFNNCFWrapperLayerMetatype
from nncf.tensorflow.graph.metatypes.matcher import get_keras_layer_metatype
from nncf.tensorflow.layers.operation import NNCFOperation
from nncf.tensorflow.layers.wrapper import NNCFWrapper

SHARED_OPERATION_MARK = "^"


def is_sequential_or_functional_model(model):
    return is_sequential_model(model) or is_functional_model(model)


def is_sequential_model(model):
    return isinstance(model, tf.keras.Sequential)


def is_functional_model(model):
    return (
        isinstance(model, tf.keras.Model)
        and not isinstance(model, tf.keras.Sequential)
        and getattr(model, "_is_graph_network", False)
    )


def is_keras_layer_model(model: tf.keras.Model) -> bool:
    """
    Checks if there is `tensorflow_hub.KerasLayer` layer in the model or not.

    :param model: Keras model to check.
    :return: `True` if there is `hub.KerasLayer` in the model and
        `False` otherwise.
    """
    for layer in model.submodules:
        if layer.__class__.__name__ == "KerasLayer":
            return True
    return False


def get_keras_layers_class_names():
    keras_layers = [
        class_name for class_name, _ in inspect.getmembers(sys.modules[tf.keras.layers.__name__], inspect.isclass)
    ]
    keras_layers += ["TensorFlowOpLayer", "TFOpLambda", "SlicingOpLambda"]
    return keras_layers


def get_keras_activation_names():
    keras_activations = [
        activation_name
        for activation_name, _ in inspect.getmembers(sys.modules[tf.keras.activations.__name__], inspect.isfunction)
    ]
    return keras_activations


def get_custom_objects(model):
    keras_layers = get_keras_layers_class_names()
    keras_activations = get_keras_activation_names()
    custom_objects = {}
    for layer in model.submodules:
        if layer.__class__.__name__ == "NNCFWrapper":
            layer = layer.layer
        if layer.__class__.__name__ not in keras_layers:
            custom_objects[layer.__class__.__name__] = layer.__class__
        if layer.__class__.__name__ == "Activation" and layer.activation.__name__ not in keras_activations:
            custom_objects[layer.activation.__name__] = layer.activation
    return custom_objects


def get_weight_name(name, layer_name=None):
    if layer_name and layer_name in name:
        return name.split(layer_name + "/")[-1]
    return name


def get_weight_by_name(layer, weight_name):
    for w in layer.weights:
        if w.name.split(":")[0] == weight_name:
            return w
    return None


def collect_wrapped_layers(model):
    wrapped_layers = []
    for layer in model.layers:
        if isinstance(layer, tf.keras.Model):
            wrapped_layers += collect_wrapped_layers(layer)
        if isinstance(layer, NNCFWrapper):
            wrapped_layers.append(layer)
    return wrapped_layers


def get_shared_node_name(layer_name: str, instance_idx: int):
    return "{}{}{}".format(layer_name, SHARED_OPERATION_MARK, instance_idx)


def get_original_name_and_instance_idx(node_name: NNCFNodeName):
    result = node_name.split(SHARED_OPERATION_MARK)
    original_name = result[0]
    instance_idx = 0 if len(result) == 1 else int(result[1])
    return original_name, instance_idx


def get_original_name(node_name):
    return get_original_name_and_instance_idx(node_name)[0]


def get_layer_to_graph_nodes_map(model, node_names):
    layer_to_nodes_map = {layer.name: {"type": layer.__class__.__name__, "nodes": []} for layer in model.layers}
    for node in node_names:
        parent_layer_name = node.split("/")[1]  # model_name/layer_name/layer_op_name/...
        if parent_layer_name not in layer_to_nodes_map:
            raise nncf.ValidationError("Could not find {} layer in Model".format(parent_layer_name))
        layer_to_nodes_map[parent_layer_name]["nodes"].append(node)
    return layer_to_nodes_map


def get_weight_node_name(graph: NNCFGraph, node_name: NNCFNodeName) -> NNCFNodeName:
    node = graph.get_node_by_name(node_name)
    while list(graph.get_previous_nodes(node)):
        node = list(graph.get_previous_nodes(node))[-1]
    return node.node_name


def get_layer_identifier(node: NNCFNode):
    layer_name, _ = get_original_name_and_instance_idx(node.node_name)
    return layer_name


def unwrap_layer(layer):
    layer_metatype = get_keras_layer_metatype(layer, determine_subtype=False)
    if layer_metatype == TFNNCFWrapperLayerMetatype:
        return layer.layer
    return layer


def get_nncf_operations(model: tf.keras.Model, operation_names: List[str]) -> Tuple[NNCFWrapper, str, NNCFOperation]:
    """
    Yields the operations from the model which names in `operation_names`.

    :param model: Wrapped model.
    :param operation_names: List of operation names.
    :return: A tuple (wrapped_layer, weight_attr, op) where
        - wrapped_layer: A wrapped layer, which contains operation weights.
        - weight_attr: A name of the attribute of the wrapped layer to which
            the operation is applied.
        - op: NNCF operation.
    """
    for wrapped_layer in collect_wrapped_layers(model):
        for weight_attr, ops in wrapped_layer.weights_attr_ops.items():
            for op in ops.values():
                if op.name in operation_names:
                    yield wrapped_layer, weight_attr, op


def _was_specially_wrapped_with_keras_export(layer, attr_name) -> bool:
    return hasattr(layer, attr_name) and getattr(layer, attr_name) != ("keras.layers.Layer",)


def is_builtin_layer(layer) -> bool:
    # A similar logic is actually what gets used in TF as
    # tensorflow.python.keras.utils.layer_utils.is_builtin_layer.
    return (
        layer.__class__.__name__ in ["SlicingOpLambda", "TFOpLambda", "TensorFlowOpLayer"]
        or _was_specially_wrapped_with_keras_export(layer, "_keras_api_names")
        or _was_specially_wrapped_with_keras_export(layer, "_keras_api_names_v1")
    )


def get_list_level(lst: List) -> int:
    return isinstance(lst, list) and list(map(get_list_level, lst or [0]))[0] + 1


def check_oplambda_input_data(x: List) -> bool:
    input_stracture = [type(item) for item in x]
    return input_stracture in ([str, int, int], [str, int, int, dict])


def reformat_inbound_nodes_for_oplambda(inbound_nodes: List) -> List[List[List]]:
    """
    Default format examples for inbound_nodes of the layer are as follows:
    1- [[['name', id, id, kwargs]]] - Single input, not shared layer.
    2- [[['name_a', id, id, kwargs], ['name_b', id, id, kwargs]]] - Multiple (two) inputs, not shared layer.
    3- [[['name_a', id, id, kwargs]], [['name_b', id, id, kwargs]]] - Single input, shared layer.

    This function assumes there are no shared layers (aka shared convolutions) among
    'TFOpLambda' or 'SlicingOpLambda' layers.

    'TFOpLambda' or 'SlicingOpLambda' layers, apart from default format templates (1 and 2 above),
    could have the following formats:
    4 - [['name', id, id, kwargs]] - Single input.
    5 - [[[['_CONSTANT_VALUE', -1, serialized_kt, {'y': ['name', id, id]}]]]] -
    Multiple (two) inputs with one constant input.

    :param inbound_nodes: Inbound nodes of the 'TFOpLambda' or 'SlicingOpLambda' layer.
    :return: Reformatted inbound_nodes.
    """
    if get_list_level(inbound_nodes) == 2:  # [[ ]] -> [[[ ]]]
        inbound_nodes = [inbound_nodes]
    if get_list_level(inbound_nodes) == 4:  # [[[[ ]]]] -> [[[ ]]]
        inbound_nodes = inbound_nodes[0]

    inbound_nodes_oplambda = []
    for inbound_node in inbound_nodes[0]:
        if inbound_node[0] != "_CONSTANT_VALUE":
            # If an element in the first call argument did not originate as a keras tensor
            # and is a constant value, it is saved using '_CONSTANT_VALUE'.
            inbound_nodes_oplambda.append(inbound_node)

        # Check for nested inbound nodes in kwargs
        kwargs = inbound_node[3]
        for item in kwargs.values():
            if isinstance(item, list) and check_oplambda_input_data(item):
                if len(item) == 3:
                    item.append({})
                inbound_nodes_oplambda.append(item)

    return [inbound_nodes_oplambda]  # convert to [[[ ]]] format, which is default for inbound nodes



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import deque
from copy import deepcopy
from typing import Any, Dict, List, Set

import networkx as nx

from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.graph import NNCFNode
from nncf.common.graph.graph import NNCFNodeName
from nncf.common.graph.graph_matching import find_subgraphs_matching_pattern
from nncf.common.graph.patterns.manager import PatternsManager
from nncf.common.graph.patterns.manager import TargetDevice
from nncf.common.utils.backend import BackendType
from nncf.common.utils.dot_file_rw import write_dot_graph
from nncf.torch.graph.graph import PTNNCFGraph
from nncf.torch.graph.operator_metatypes import PTRELUMetatype


class SearchGraphNode:
    """
    Class describing nodes used in SearchGraph.
    """

    def __init__(self, node_key: str, data: Dict):
        self.node_key = node_key
        self.data = data if data else {}

    @property
    def is_merged(self) -> bool:
        """
        Returns value that the node is merged.
        """
        return self.data.get(SearchGraph.IS_MERGED_NODE_ATTR)

    @property
    def is_dummy(self) -> bool:
        """
        Returns value that the node is dummy.
        """
        return self.data.get(SearchGraph.IS_DUMMY_NODE_ATTR)

    @property
    def node_name(self) -> NNCFNodeName:
        return self.data.get(NNCFNode.NODE_NAME_ATTR)

    @property
    def node_type(self) -> str:
        """
        Returns type of node.
        """
        return self.data.get(NNCFNode.NODE_TYPE_ATTR)

    @property
    def layer_name(self) -> str:
        """
        Returns the name of the layer to which the node corresponds.
        """
        return self.data.get(NNCFNode.LAYER_NAME_ATTR)

    @property
    def main_id(self) -> int:
        """
        Returns the id of node. In case if the node is merged returns id of the first node from merged node list.
        """
        if not self.is_merged:
            return self.data.get("id")
        return self.data.get(SearchGraph.MERGED_NODES_NODE_ATTR)[0].get("id")

    @property
    def bottom_id(self) -> int:
        """
        Returns the id of node. In case if the node is merged returns id of the last node from merged node list.
        """
        if not self.is_merged:
            return self.data.get("id")
        return self.data.get(SearchGraph.MERGED_NODES_NODE_ATTR)[-1].get("id")

    def __str__(self):
        return self.node_key

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return isinstance(other, SearchGraphNode) and self.node_key == other.node_key


ShapeVsNodesMap = Dict[str, Set[SearchGraphNode]]


class SearchGraph:
    """
    A wrapper over the graph, which represents the DNN execution graph transformed
    by pattern matching, merging nodes and inserting auxiliary nodes.
    """

    ACTIVATION_OUTPUT_SHAPE_ATTR = "activation_output_shape"
    IS_MERGED_NODE_ATTR = "is_merged"
    IS_DUMMY_NODE_ATTR = "is_dummy"
    KEY_NODE_ATTR = "key"
    MERGED_NODES_NODE_ATTR = "merged_nodes"
    TYPE_NODE_ATTR = "type"
    DUMMY_POSTFIX = " dummy"

    def __init__(self, nx_merged_graph: nx.DiGraph):
        self._nx_graph = nx_merged_graph
        old_merged_graph_nodes = deepcopy(self._nx_graph._node)
        for node_key in old_merged_graph_nodes:
            next_nodes = self._nx_graph._succ[node_key]
            if len(list(next_nodes)) > 1:
                self._insert_dummy_node(node_key)

    def _nx_node_to_sgraph_node(self, nx_node_key: str, nx_node_attrs: Dict[str, Any]):
        return SearchGraphNode(nx_node_key, nx_node_attrs)

    def get_node_by_key(self, node_key: str) -> SearchGraphNode:
        """
        :param node_key: key (node_name) of the node.
        :return: SearchGraphNode in a graph with such key.
        """
        return SearchGraphNode(node_key, self._nx_graph.nodes[node_key])

    def get_all_nodes(self) -> List[SearchGraphNode]:
        """
        Returns list of all graph nodes.
        """
        all_nodes = []
        for node_key, node_attrs in self._nx_graph.nodes.items():
            all_nodes.append(SearchGraphNode(node_key, node_attrs))
        return all_nodes

    def get_next_nodes(self, node_key: str) -> List[SearchGraphNode]:
        """
        Returns consumer nodes of provided node key.

        :param node_key: Producer node key.
        :return: List of consumer nodes of provided node.
        """
        next_node_keys = self._nx_graph.succ[node_key]
        return [self.get_node_by_key(node_key) for node_key in next_node_keys]

    def get_previous_nodes(self, node: SearchGraphNode) -> List[SearchGraphNode]:
        """
        Returns producer nodes of provided node.

        :param node: Consumer node.
        :return: List of producers nodes of provided node.
        """
        nx_node_keys = self._nx_graph.pred[node.node_key]
        return [self.get_node_by_key(node_key) for node_key in nx_node_keys]

    def set_node_attr(self, node_key: str, name_attr: str, value_attr: str):
        """
        Set value of attribute by name for a given node with the same key.
        """
        self._nx_graph.nodes[node_key][name_attr] = value_attr

    def get_prev_nodes(self, node_key: str) -> List[SearchGraphNode]:
        """
        Returns producer nodes of provided node key.

        :param node_key: Consumer node key.
        :return: List of producers nodes of provided node.
        """
        prev_node_keys = self._nx_graph.pred[node_key]
        return [self.get_node_by_key(node_key) for node_key in prev_node_keys]

    def get_prev_edges(self, node_key: str) -> Dict[str, Any]:
        return self._nx_graph.pred[node_key]

    def get_next_edges(self, node_key: str) -> Dict[str, Any]:
        return self._nx_graph.succ[node_key]

    def _insert_dummy_node(self, node_key: str):
        next_nodes = deepcopy(self._nx_graph._succ[node_key])
        dummy_node_attr = {SearchGraph.IS_DUMMY_NODE_ATTR: True}
        node_key_attr = deepcopy(self._nx_graph._node[node_key])
        dummy_node_key = node_key + SearchGraph.DUMMY_POSTFIX
        node_key_attr.update(dummy_node_attr)
        self._nx_graph.add_node(dummy_node_key, **node_key_attr)

        edge_attrs = None
        for next_node_key in next_nodes:
            edge_attrs = self._nx_graph.get_edge_data(node_key, next_node_key)
            self._nx_graph.remove_edge(node_key, next_node_key)
            self._nx_graph.add_edge(dummy_node_key, next_node_key, **edge_attrs)
        self._nx_graph.add_edge(node_key, dummy_node_key, **edge_attrs)

    def get_nx_graph(self) -> nx.DiGraph:
        """
        Returns internal representation of SearchGraph as a networkx graph.
        """
        return self._nx_graph

    def _get_graph_for_visualization(self) -> nx.DiGraph:
        """
        :return: A user-friendly graph .dot file, making it easier to debug the network and setup
        ignored/target scopes.
        """
        out_graph = nx.DiGraph()
        for node in self.get_all_nodes():
            attrs_node = {}
            if node.is_merged:
                attrs_node["label"] = f"main: {node.main_id} bottom: {node.bottom_id} {node.node_key}"
            elif node.is_dummy:
                attrs_node["label"] = f"dummy {node.node_key}"
            else:
                attrs_node["label"] = f"id: {node.node_key}"
            out_graph.add_node(node.node_key, **attrs_node)

        for u, v in self._nx_graph.edges:
            edge = self._nx_graph.edges[u, v]
            style = "solid"
            out_graph.add_edge(u, v, label=edge[NNCFGraph.ACTIVATION_SHAPE_EDGE_ATTR], style=style)

        mapping = {k: v["label"] for k, v in out_graph.nodes.items()}
        out_graph = nx.relabel_nodes(out_graph, mapping)
        for node in out_graph.nodes.values():
            node.pop("label")

        return out_graph

    def visualize_graph(self, path: str):
        out_graph = self._get_graph_for_visualization()
        write_dot_graph(out_graph, path)


def get_search_graph(original_graph: PTNNCFGraph, hw_fused_ops: bool) -> SearchGraph:
    """
    Returns a transformed representation of the network graph for blocks searching.
    """
    nx_merged_graph = get_merged_original_graph_with_pattern(original_graph.get_nx_graph_copy(), hw_fused_ops)
    sgraph = SearchGraph(nx_merged_graph)
    return sgraph


def get_merged_original_graph_with_pattern(orig_graph: nx.DiGraph, hw_fused_ops: bool) -> nx.DiGraph:
    """
    :param orig_graph: Original graph of model
    :param hw_fused_ops: indicates whether to merge operations by hw fusing pattern. Merged operation can't be part
    of different skipping block.
    :return: Graph with merged nodes by patterns
    """
    merged_graph = orig_graph
    if not hw_fused_ops:
        return merged_graph

    pattern_fusing_graph = PatternsManager.get_full_hw_pattern_graph(backend=BackendType.TORCH, device=TargetDevice.ANY)
    matches = find_subgraphs_matching_pattern(orig_graph, pattern_fusing_graph)
    nx.set_node_attributes(merged_graph, False, SearchGraph.IS_DUMMY_NODE_ATTR)
    nx.set_node_attributes(merged_graph, False, SearchGraph.IS_MERGED_NODE_ATTR)
    for match in matches:
        if len(match) == 1:
            continue
        input_node_key = match[0]
        output_node_key = match[-1]
        in_edges = list(merged_graph.in_edges(input_node_key))
        out_edges = list(merged_graph.out_edges(output_node_key))

        in_edge_copies_dict = {}
        for in_edge_key in in_edges:
            in_edge_copies_dict[in_edge_key] = deepcopy(merged_graph.edges[in_edge_key])
        out_edge_copies_dict = {}
        for out_edge_key in out_edges:
            out_edge_copies_dict[out_edge_key] = deepcopy(merged_graph.edges[out_edge_key])

        merged_node_key = ""
        merged_nodes = []
        type_list = []
        for node_key in match:
            attrs = orig_graph.nodes[node_key]
            merged_node_key += str(attrs["id"]) + " " + attrs[SearchGraph.TYPE_NODE_ATTR] + "  "

            merged_nodes.append(orig_graph.nodes[node_key])
            merged_graph.remove_node(node_key)
            type_list.append(attrs[SearchGraph.TYPE_NODE_ATTR])
        merged_node_attrs = {
            SearchGraph.KEY_NODE_ATTR: merged_node_key,
            SearchGraph.IS_MERGED_NODE_ATTR: True,
            SearchGraph.TYPE_NODE_ATTR: type_list,
            SearchGraph.MERGED_NODES_NODE_ATTR: merged_nodes,
        }
        merged_graph.add_node(merged_node_key, **merged_node_attrs)
        for in_edge_key, in_edge_attrs in in_edge_copies_dict.items():
            merged_graph.add_edge(in_edge_key[0], merged_node_key, **in_edge_attrs)
        for out_edge_key, out_edge_attrs in out_edge_copies_dict.items():
            merged_graph.add_edge(merged_node_key, out_edge_key[1], **out_edge_attrs)

    return merged_graph


def check_graph_has_no_hanging_edges_after_block_removal(
    graph: SearchGraph, first_skipped_node: SearchGraphNode, end_node: SearchGraphNode
) -> bool:
    """
    The subgraph is traversed starting with the first_skipped_node and ending with the end_node
    to determine that after deleting such a block there are no dangling edges in the graph.
    """
    #            A
    #           / \
    #          /   \
    #         /     \
    #    first_skipped_node  |   Edge A-C is dangling edges
    #        |       |
    #         \     /
    #          \   /
    #            C
    #            |
    #           ...
    #            |
    #         end_node
    start_node = first_skipped_node
    if not first_skipped_node.is_dummy:
        previous_nodes = graph.get_previous_nodes(first_skipped_node)
        num_inputs = len(previous_nodes)
        assert num_inputs == 1, f"building block should have a single input, but it has {num_inputs} inputs."
        start_node = previous_nodes[0]
    q = deque([start_node])
    addit_nodes = set()
    nodes = []
    potential_end_nodes = []
    while len(q) != 0:
        current_node = q.pop()
        if current_node.main_id != start_node.main_id:
            prev_nodes = graph.get_prev_nodes(current_node.node_key)
            if len(prev_nodes) > 1:
                for pn in prev_nodes:
                    if pn.bottom_id < start_node.bottom_id:
                        return False  # there is extra edge
                    addit_nodes.add(pn)
        if current_node.node_key == end_node.node_key:
            continue
        if current_node.main_id > end_node.main_id:
            return False
        if current_node not in nodes:
            nodes.append(current_node)
            next_nodes = graph.get_next_nodes(current_node.node_key)
            if len(next_nodes) == 0:
                potential_end_nodes.append(current_node)
            for next_node in next_nodes:
                q.appendleft(next_node)
    if len(q) > 0 or len(potential_end_nodes) > 0:
        return False
    for node in addit_nodes:
        if node not in nodes:
            return False
    nodes.append(end_node)
    return True


def check_graph_has_no_duplicate_edges_after_block_removal(
    sgraph: SearchGraph, first_skipped_node: SearchGraphNode, end_node: SearchGraphNode
) -> bool:
    """
    This rule ensures that no duplicate edges will be created in the graph after a block is deleted.
    """
    #         A              A
    #        / \            / \
    #       /   \          /   \
    #    block   |  =>     \   /
    #       \   /           \ /
    #        \ /             D
    #         D          forbidden

    if first_skipped_node.is_dummy:
        next_end_node = sgraph.get_next_nodes(end_node.node_key)
        if len(next_end_node) != 0:
            attr = sgraph._nx_graph.get_edge_data(first_skipped_node.node_key, next_end_node[0].node_key)
        else:
            attr = None
    else:
        previous_node = sgraph.get_prev_nodes(first_skipped_node.node_key)
        next_end_node = sgraph.get_next_nodes(end_node.node_key)
        if len(previous_node) != 0 and len(next_end_node) != 0:
            attr = sgraph._nx_graph.get_edge_data(previous_node[0].node_key, next_end_node[0].node_key)
        else:
            attr = None
    return attr is None


def check_graph_has_no_act_layer_duplication_after_block_removal(
    sgraph: SearchGraph, first_skipped_node: SearchGraphNode, end_node: SearchGraphNode
) -> bool:
    """
    This rule ensures that after the block is deleted there will be no duplication of activation layers.
    """
    #         A             A
    #         |             |
    #        relu          relu
    #         |      =>     |
    #        block         relu
    #         |
    #        relu        forbidden

    previous_nodes = sgraph.get_prev_nodes(first_skipped_node.node_key)
    next_end_node = sgraph.get_next_nodes(end_node.node_key)
    if len(next_end_node) == 0 or len(previous_nodes) == 0:
        return True
    if previous_nodes[0].is_dummy:
        previous_nodes = sgraph.get_prev_nodes(previous_nodes[0].node_key)

    if (
        previous_nodes[0].node_type[-1] in PTRELUMetatype.get_all_aliases()
        and next_end_node[0].node_type[0] in PTRELUMetatype.get_all_aliases()
    ):
        return False
    return True


def get_num_ops_in_block(first_skipped_node: SearchGraphNode, end_node: SearchGraphNode) -> int:
    """
    Calculates number of operations in the block by using indexes. Indexes should be in execution order.
    The block is defined by first and last skipped node.
    """
    num_ops = end_node.bottom_id - first_skipped_node.main_id
    if first_skipped_node.is_dummy:
        num_ops -= 1
    return num_ops



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect
from collections import OrderedDict
from typing import Any, Dict, List, Optional
from typing import OrderedDict as OrderedDictType
from typing import Tuple

from nncf.common.logging import nncf_logger
from nncf.common.pruning.weights_flops_calculator import WeightsFlopsCalculator
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.base_handler import ElasticityConfig
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.base_handler import ElasticityHandler
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.base_handler import SingleElasticityHandler
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.elastic_depth import ElasticDepthHandler
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.elastic_depth import ElasticDepthSearchSpace
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.elastic_kernel import ElasticKernelHandler
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.elastic_kernel import ElasticKernelSearchSpace
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.elastic_width import ElasticWidthHandler
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.elastic_width import ElasticWidthSearchSpace
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.elasticity_dim import ElasticityDim
from nncf.torch.graph.operator_metatypes import PTModuleConv1dMetatype
from nncf.torch.graph.operator_metatypes import PTModuleConv2dMetatype
from nncf.torch.graph.operator_metatypes import PTModuleConv3dMetatype
from nncf.torch.graph.operator_metatypes import PTModuleConvTranspose2dMetatype
from nncf.torch.graph.operator_metatypes import PTModuleConvTranspose3dMetatype
from nncf.torch.graph.operator_metatypes import PTModuleDepthwiseConv1dSubtype
from nncf.torch.graph.operator_metatypes import PTModuleDepthwiseConv2dSubtype
from nncf.torch.graph.operator_metatypes import PTModuleDepthwiseConv3dSubtype
from nncf.torch.graph.operator_metatypes import PTModuleLinearMetatype
from nncf.torch.nncf_network import NNCFNetwork
from nncf.torch.pruning.utils import collect_output_shapes

SubnetConfig = OrderedDictType[ElasticityDim, ElasticityConfig]


class MEHandlerStateNames:
    IS_HANDLER_ENABLED_MAP = "is_handler_enabled_map"
    STATES_OF_HANDLERS = "states_of_handlers"


class MultiElasticityHandler(ElasticityHandler):
    """
    An interface for handling multiple elasticity in the network. The elasticity defines variable values in properties
    of the layers or the network, e.g. variable number of channels in the Conv or variable number of layers in the
    network. By applying elasticity it's possible to derive a smaller models (Subnets) that have some elements in
    common with the original model.
    The interface defines methods for activation Subnets.
    """

    _state_names = MEHandlerStateNames

    def __init__(self, handlers: OrderedDictType[ElasticityDim, SingleElasticityHandler], target_model: NNCFNetwork):
        GENERAL_CONV_LAYER_METATYPES = [
            PTModuleConv1dMetatype,
            PTModuleDepthwiseConv1dSubtype,
            PTModuleConv2dMetatype,
            PTModuleDepthwiseConv2dSubtype,
            PTModuleConv3dMetatype,
            PTModuleDepthwiseConv3dSubtype,
            PTModuleConvTranspose2dMetatype,
            PTModuleConvTranspose3dMetatype,
        ]
        LINEAR_LAYER_METATYPES = [PTModuleLinearMetatype]
        self._handlers = handlers
        self._target_model = target_model
        self._is_handler_enabled_map = {elasticity_dim: True for elasticity_dim in handlers}
        self._weights_calc = WeightsFlopsCalculator(
            conv_op_metatypes=GENERAL_CONV_LAYER_METATYPES, linear_op_metatypes=LINEAR_LAYER_METATYPES
        )
        self.activate_supernet()

    @property
    def width_search_space(self) -> ElasticWidthSearchSpace:
        return self.width_handler.get_search_space()

    @property
    def kernel_search_space(self) -> ElasticKernelSearchSpace:
        return self.kernel_handler.get_search_space()

    @property
    def depth_search_space(self) -> ElasticDepthSearchSpace:
        return self.depth_handler.get_search_space()

    @property
    def width_handler(self) -> Optional[ElasticWidthHandler]:
        return self._get_handler_by_elasticity_dim(ElasticityDim.WIDTH)

    @property
    def kernel_handler(self) -> Optional[ElasticKernelHandler]:
        return self._get_handler_by_elasticity_dim(ElasticityDim.KERNEL)

    @property
    def depth_handler(self) -> Optional[ElasticDepthHandler]:
        return self._get_handler_by_elasticity_dim(ElasticityDim.DEPTH)

    def get_available_elasticity_dims(self) -> List[ElasticityDim]:
        """
        :return: list of available elasticity dimension. E.g. it's possible to have a single elasticity like Elastic
        Depth or all possible elasticities (Elastic Depth/Width/Kernel) added to the model.
        """
        return list(self._handlers)

    def get_active_config(self) -> SubnetConfig:
        """
        Forms an elasticity configuration that describes currently activated Subnet across all elasticities.

        :return: elasticity configuration
        """
        return self._collect_handler_data_by_method_name(self._get_current_method_name())

    def get_random_config(self) -> SubnetConfig:
        """
        Forms an elasticity configuration that describes a Subnet with randomly chosen elastic values across all
        elasticities.

        :return: elasticity configuration
        """
        return self._collect_handler_data_by_method_name(self._get_current_method_name())

    def get_minimum_config(self) -> SubnetConfig:
        """
        Forms an elasticity configuration that describes a Subnet with minimum elastic values across all elasticities.

        :return: elasticity configuration
        """
        return self._collect_handler_data_by_method_name(self._get_current_method_name())

    def get_maximum_config(self) -> SubnetConfig:
        """
        Forms an elasticity configuration that describes a Subnet with maximum elastic values across all elasticities.

        :return: elasticity configuration
        """
        return self._collect_handler_data_by_method_name(self._get_current_method_name())

    def activate_supernet(self) -> None:
        """
        Activates the Supernet - the original network to which elasticity was applied.
        """
        self._collect_handler_data_by_method_name(self._get_current_method_name())

    def activate_subnet_for_config(self, config: SubnetConfig) -> None:
        """
        Activates a Subnet that corresponds to the given elasticity configuration

        :param config: elasticity configuration
        """
        active_handlers = {dim: self._handlers[dim] for dim in self._handlers if self._is_handler_enabled_map[dim]}
        for handler_id, handler in self._handlers.items():
            if handler_id in config:
                sub_config = config[handler_id]
                other_active_handlers = dict(filter(lambda pair: pair[0] != handler_id, active_handlers.items()))
                resolved_config = handler.resolve_conflicts_with_other_elasticities(sub_config, other_active_handlers)
                handler.activate_subnet_for_config(resolved_config)
                if sub_config != resolved_config:
                    nncf_logger.warning(
                        f"Config for {handler_id} mismatch. Requested: {sub_config}. Resolved: {resolved_config}"
                    )

    def load_state(self, state: Dict[str, Any]) -> None:
        """
        Initializes object from the state.

        :param state: Output of `get_state()` method.
        """
        states_of_handlers = state[self._state_names.STATES_OF_HANDLERS]
        is_handler_enabled_map = state[self._state_names.IS_HANDLER_ENABLED_MAP]

        for dim_str, handler_state in states_of_handlers.items():
            dim = ElasticityDim(dim_str)
            if dim in self._handlers:
                self._handlers[dim].load_state(handler_state)

        for dim_str, is_enabled in is_handler_enabled_map.items():
            dim = ElasticityDim(dim_str)
            self._is_handler_enabled_map[dim] = is_enabled

    def get_state(self) -> Dict[str, Any]:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.

        :return: state of the object
        """
        states_of_handlers = {dim.value: handler.get_state() for dim, handler in self._handlers.items()}
        is_handler_enabled_map = {dim.value: is_enabled for dim, is_enabled in self._is_handler_enabled_map.items()}
        return {
            self._state_names.STATES_OF_HANDLERS: states_of_handlers,
            self._state_names.IS_HANDLER_ENABLED_MAP: is_handler_enabled_map,
        }

    def get_search_space(self) -> Dict[str, Any]:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents the search space of the super-network.

        :return: search space
        """
        active_handlers = {dim: self._handlers[dim] for dim in self._handlers if self._is_handler_enabled_map[dim]}
        space = {}
        for handler_id, handler in active_handlers.items():
            space[handler_id.value] = handler.get_search_space()
        return space

    def enable_all(self) -> None:
        """
        Enables all elasticities for being selected on sampling subnets.
        """
        self._is_handler_enabled_map = {elasticity_dim: True for elasticity_dim in self._is_handler_enabled_map}

    def disable_all(self) -> None:
        """
        Disables all elasticities for being selected on sampling subnets.
        """
        self._is_handler_enabled_map = {elasticity_dim: False for elasticity_dim in self._is_handler_enabled_map}

    def enable_elasticity(self, dim: ElasticityDim):
        """
        Enables elasticity for being selected on sampling subnets by a given type.
        """
        self._is_handler_enabled_map[dim] = True

    def disable_elasticity(self, dim: ElasticityDim):
        """
        Disables elasticity for being selected on sampling subnets by a given type.
        """
        self._is_handler_enabled_map[dim] = False

    def count_flops_and_weights_for_active_subnet(self) -> Tuple[int, int]:
        """
        :return: FLOPs and the number weights and in the model for convolution and fully connected layers.
        """
        kernel_sizes = None
        if self.kernel_handler is not None:
            kernel_sizes = self.kernel_handler.get_active_kernel_sizes_per_node()

        names_of_skipped_nodes = None
        if self.depth_handler is not None:
            names_of_skipped_nodes = self.depth_handler.get_names_of_skipped_nodes()

        input_width_values, output_width_values = None, None
        if self.width_handler is not None:
            input_width_values, output_width_values = self.width_handler.get_active_in_out_width_values()

        graph = self._target_model.nncf.get_graph()
        output_shapes = collect_output_shapes(graph)

        flops, num_weights = self._weights_calc.count_flops_and_weights(
            graph=graph,
            output_shapes=output_shapes,
            input_channels=input_width_values,
            output_channels=output_width_values,
            kernel_sizes=kernel_sizes,
            op_addresses_to_skip=names_of_skipped_nodes,
        )

        return flops, num_weights

    def _get_handler_by_elasticity_dim(self, dim: ElasticityDim) -> Optional[SingleElasticityHandler]:
        result = None
        if dim in self._handlers:
            result = self._handlers[dim]
        return result

    def _collect_handler_data_by_method_name(self, method_name: str) -> OrderedDictType[ElasticityDim, Any]:
        result = OrderedDict()
        for elasticity_dim, handler in self._handlers.items():
            if self._is_handler_enabled_map[elasticity_dim]:
                handler_method = getattr(handler, method_name)
                data = handler_method()
                result[elasticity_dim] = data
        return result

    @staticmethod
    def _get_current_method_name() -> str:
        return inspect.stack()[1].function

    def get_design_vars_info(self) -> float:
        active_handlers = {dim: self._handlers[dim] for dim in self._handlers if self._is_handler_enabled_map[dim]}
        num_vars = 0
        vars_upper = []
        for handler_id, handler in active_handlers.items():
            if handler_id is ElasticityDim.DEPTH:
                num_vars += 1
                vars_upper.append(len(handler.get_search_space()) - 1)
            else:
                num_vars += len(handler.get_search_space())
                vars_upper += [len(handler.get_search_space()[i]) - 1 for i in range(len(handler.get_search_space()))]
        return num_vars, vars_upper

    def get_config_from_pymoo(self, x: List) -> SubnetConfig:
        active_handlers = {dim: self._handlers[dim] for dim in self._handlers if self._is_handler_enabled_map[dim]}
        index_pos = 0
        sample = SubnetConfig()
        for handler_id, _ in active_handlers.items():
            if handler_id is ElasticityDim.KERNEL:
                sample[handler_id] = [
                    self.kernel_search_space[j - index_pos][x[j]]
                    for j in range(index_pos, index_pos + len(self.kernel_search_space))
                ]
                index_pos += len(self.kernel_search_space)
            elif handler_id is ElasticityDim.WIDTH:
                sample[handler_id] = {
                    key - index_pos: self.width_search_space[key - index_pos][x[key]]
                    for key in range(index_pos, index_pos + len(self.width_search_space))
                }
                index_pos += len(self.width_search_space)
            elif handler_id is ElasticityDim.DEPTH:
                sample[handler_id] = self.depth_search_space[x[index_pos]]
                index_pos += 1
        return sample



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional, Tuple

import numpy as np
import onnx

from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.graph.transformations.commands import TargetType
from nncf.experimental.common.tensor_statistics.collectors import TensorCollector
from nncf.onnx.graph.model_utils import remove_fq_from_inputs
from nncf.onnx.graph.node_utils import get_bias_value
from nncf.onnx.graph.node_utils import is_any_weight_quantized
from nncf.onnx.graph.node_utils import is_node_with_bias
from nncf.onnx.graph.onnx_helper import get_name_to_node_map
from nncf.onnx.graph.transformations.command_creation import create_bias_correction_command
from nncf.onnx.graph.transformations.commands import ONNXInitializerUpdateCommand
from nncf.onnx.graph.transformations.commands import ONNXModelExtractionCommand
from nncf.onnx.graph.transformations.commands import ONNXOutputInsertionCommand
from nncf.onnx.graph.transformations.commands import ONNXTargetPoint
from nncf.onnx.statistics.collectors import ONNXNNCFCollectorTensorProcessor
from nncf.onnx.statistics.collectors import get_mean_statistic_collector
from nncf.onnx.statistics.collectors import get_raw_stat_collector
from nncf.onnx.tensor import ONNXNNCFTensor
from nncf.quantization.algorithms.bias_correction.backend import BiasCorrectionAlgoBackend


class ONNXBiasCorrectionAlgoBackend(BiasCorrectionAlgoBackend):
    @property
    def tensor_processor(self) -> ONNXNNCFCollectorTensorProcessor:
        return ONNXNNCFCollectorTensorProcessor

    @staticmethod
    def target_point(target_type: TargetType, target_node_name: str, port_id: int) -> ONNXTargetPoint:
        return ONNXTargetPoint(target_type, target_node_name, port_id)

    @staticmethod
    def create_bias_correction_command(
        node: NNCFNode, bias_value: np.ndarray, nncf_graph: NNCFGraph
    ) -> ONNXInitializerUpdateCommand:
        return create_bias_correction_command(node, bias_value)

    @staticmethod
    def model_extraction_command(
        input_ids: List[Tuple[str, int]], output_ids: List[Tuple[str, int]]
    ) -> ONNXModelExtractionCommand:
        return ONNXModelExtractionCommand(input_ids, output_ids)

    @staticmethod
    def output_insertion_command(nncf_graph: NNCFGraph, target_point: ONNXTargetPoint) -> ONNXOutputInsertionCommand:
        nncf_input_node_next_nodes = {}
        for input_node in nncf_graph.get_input_nodes():
            next_nodes = nncf_graph.get_next_nodes(input_node)
            nncf_input_node_next_nodes[input_node.node_name] = [node.node_name for node in next_nodes]
        return ONNXOutputInsertionCommand(target_point, nncf_input_node_next_nodes)

    @staticmethod
    def mean_statistic_collector(
        channel_axis: int,
        inplace: bool,
        num_samples: Optional[int] = None,
        window_size: Optional[int] = None,
    ) -> TensorCollector:
        return get_mean_statistic_collector(num_samples, channel_axis, window_size, inplace)

    @staticmethod
    def raw_statistic_collector(num_samples: int = None) -> TensorCollector:
        return get_raw_stat_collector(num_samples)

    @staticmethod
    def process_model_output(raw_data: Dict, output_name: str) -> ONNXNNCFTensor:
        return ONNXNNCFTensor(raw_data[output_name])

    @staticmethod
    def get_activation_port_id(node: NNCFNode, nncf_graph: NNCFGraph) -> Tuple[int, int]:
        return 0

    @staticmethod
    def get_bias_value(node: NNCFNode, model: onnx.ModelProto, nncf_graph: NNCFGraph) -> np.ndarray:
        return get_bias_value(node, model)

    @staticmethod
    def get_input_name(model: onnx.ModelProto, node_name: str, input_port_id: int) -> str:
        node_mapping = get_name_to_node_map(model)
        return node_mapping[node_name].input[input_port_id]

    @staticmethod
    def get_output_name(model: onnx.ModelProto, node_name: str, output_port_id: int) -> str:
        node_mapping = get_name_to_node_map(model)
        return node_mapping[node_name].output[output_port_id]

    @staticmethod
    def is_quantized_weights(node: NNCFNode, nncf_graph: NNCFGraph) -> bool:
        return is_any_weight_quantized(node, nncf_graph)

    @staticmethod
    def is_node_with_bias(node: NNCFNode, nncf_graph: NNCFGraph) -> bool:
        return is_node_with_bias(node)

    @staticmethod
    def remove_fq_from_inputs(model: onnx.ModelProto, nncf_graph: NNCFGraph) -> onnx.ModelProto:
        return remove_fq_from_inputs(model, nncf_graph)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nncf.config.definitions import FILTER_PRUNING_ALGO_NAME_IN_CONFIG
from nncf.config.definitions import ONLINE_DOCS_ROOT
from nncf.config.schemata.basic import BOOLEAN
from nncf.config.schemata.basic import NUMBER
from nncf.config.schemata.basic import STRING
from nncf.config.schemata.basic import with_attributes
from nncf.config.schemata.common.compression import BASIC_COMPRESSION_ALGO_SCHEMA
from nncf.config.schemata.common.targeting import GENERIC_INITIALIZER_SCHEMA
from nncf.config.schemata.common.targeting import SCOPING_PROPERTIES
from nncf.config.schemata.defaults import PRUNE_BATCH_NORMS
from nncf.config.schemata.defaults import PRUNE_DOWNSAMPLE_CONVS
from nncf.config.schemata.defaults import PRUNE_FIRST_CONV
from nncf.config.schemata.defaults import PRUNING_ALL_WEIGHTS
from nncf.config.schemata.defaults import PRUNING_FILTER_IMPORTANCE
from nncf.config.schemata.defaults import PRUNING_INIT
from nncf.config.schemata.defaults import PRUNING_INTERLAYER_RANKING_TYPE
from nncf.config.schemata.defaults import PRUNING_LEGR_GENERATIONS
from nncf.config.schemata.defaults import PRUNING_LEGR_MAX_PRUNING
from nncf.config.schemata.defaults import PRUNING_LEGR_MUTATE_PERCENT
from nncf.config.schemata.defaults import PRUNING_LEGR_NUM_SAMPLES
from nncf.config.schemata.defaults import PRUNING_LEGR_POPULATION_SIZE
from nncf.config.schemata.defaults import PRUNING_LEGR_RANDOM_SEED
from nncf.config.schemata.defaults import PRUNING_LEGR_SIGMA_SCALE
from nncf.config.schemata.defaults import PRUNING_LEGR_TRAIN_STEPS
from nncf.config.schemata.defaults import PRUNING_NUM_INIT_STEPS
from nncf.config.schemata.defaults import PRUNING_SCHEDULE
from nncf.config.schemata.defaults import PRUNING_STEPS
from nncf.config.schemata.defaults import PRUNING_TARGET

FILTER_PRUNING_SCHEDULE_OPTIONS = ["exponential", "exponential_with_bias", "baseline"]
FILTER_IMPORTANCE_OPTIONS = ["L2", "L1", "geometric_median"]
INTERLAYER_RANKING_TYPE_OPTIONS = ["unweighted_ranking", "learned_ranking"]

FILTER_PRUNING_SCHEMA = {
    **BASIC_COMPRESSION_ALGO_SCHEMA,
    "description": f"Applies filter pruning during training of the model to effectively remove entire "
    f"sub-dimensions of tensors in the original model from computation and therefore increase "
    f"performance.\n"
    f"See [Pruning.md]"
    f"({ONLINE_DOCS_ROOT}"
    f"/docs/compression_algorithms/Pruning.md) and the rest of this schema for "
    f"more details and parameters.",
    "properties": {
        "algorithm": {"const": FILTER_PRUNING_ALGO_NAME_IN_CONFIG},
        "initializer": GENERIC_INITIALIZER_SCHEMA,
        "pruning_init": with_attributes(
            NUMBER,
            description="Initial value of the pruning level applied to the prunable operations.",
            default=PRUNING_INIT,
        ),
        "params": {
            "type": "object",
            "properties": {
                "filter_importance": with_attributes(
                    STRING,
                    description="The type of filter importance metric.",
                    enum=FILTER_IMPORTANCE_OPTIONS,
                    default=PRUNING_FILTER_IMPORTANCE,
                ),
                "pruning_target": with_attributes(
                    NUMBER,
                    description="Target value of the pruning level for "
                    "the operations that can be pruned. "
                    "The operations are determined by analysis of the "
                    "model architecture during the pruning algorithm "
                    "initialization stage.",
                    default=PRUNING_TARGET,
                ),
                "pruning_steps": with_attributes(
                    NUMBER,
                    description="Number of epochs during which the pruning level is "
                    "increased from `pruning_init` to `pruning_target`.",
                    default=PRUNING_STEPS,
                ),
                "pruning_flops_target": with_attributes(
                    NUMBER, description="Target value of the pruning level for model FLOPs."
                ),
                "schedule": with_attributes(
                    STRING,
                    description="The type of scheduling to use for adjusting the target pruning level.",
                    enum=FILTER_PRUNING_SCHEDULE_OPTIONS,
                    default=PRUNING_SCHEDULE,
                ),
                "num_init_steps": with_attributes(
                    NUMBER,
                    description="Number of epochs for model pretraining before starting filter pruning.",
                    default=PRUNING_NUM_INIT_STEPS,
                ),
                "interlayer_ranking_type": with_attributes(
                    STRING,
                    description="The type of filter ranking across the layers.",
                    enum=INTERLAYER_RANKING_TYPE_OPTIONS,
                    default=PRUNING_INTERLAYER_RANKING_TYPE,
                ),
                "all_weights": with_attributes(
                    BOOLEAN,
                    description="Whether to prune layers independently (choose filters "
                    "with the smallest importance in each layer separately) "
                    "or not.",
                    default=PRUNING_ALL_WEIGHTS,
                ),
                "prune_first_conv": with_attributes(
                    BOOLEAN,
                    description="Whether to prune first convolutional layers or "
                    "not. A 'first' convolutional layer is such a "
                    "layer that the path from model input to "
                    "this layer has no other "
                    "convolution operations on it.",
                    default=PRUNE_FIRST_CONV,
                ),
                "prune_downsample_convs": with_attributes(
                    BOOLEAN,
                    description="Whether to prune downsampling convolutional layers (with stride > 1) or not.",
                    default=PRUNE_DOWNSAMPLE_CONVS,
                ),
                "prune_batch_norms": with_attributes(
                    BOOLEAN,
                    description="Whether to prune parameters of the Batch Norm "
                    "layer that corresponds to pruned filters of the "
                    "convolutional layer which feeds its output to "
                    "this Batch Norm.",
                    default=PRUNE_BATCH_NORMS,
                ),
                "legr_params": {
                    "type": "object",
                    "description": f"Describes parameters specific to the LeGR pruning algorithm."
                    f"See [Pruning.md]"
                    f"({ONLINE_DOCS_ROOT}"
                    f"/docs/compression_algorithms/Pruning.md#interlayer-ranking-types) "
                    f"for more details.",
                    "properties": {
                        "generations": with_attributes(
                            NUMBER,
                            description="Number of generations for the evolution algorithm.",
                            default=PRUNING_LEGR_GENERATIONS,
                        ),
                        "train_steps": with_attributes(
                            NUMBER,
                            description="Number of training steps to estimate pruned model accuracy.",
                            default=PRUNING_LEGR_TRAIN_STEPS,
                        ),
                        "max_pruning": with_attributes(
                            NUMBER,
                            description="Pruning level for the model to train "
                            "LeGR algorithm on it. If learned ranking "
                            "will be used for multiple pruning "
                            "levels, the highest should be used as "
                            "`max_pruning`. If model will be pruned "
                            "with one pruning level, this target should "
                            "be used.",
                            default=PRUNING_LEGR_MAX_PRUNING,
                        ),
                        "random_seed": with_attributes(
                            NUMBER,
                            description="Random seed for LeGR coefficients generation.",
                            default=PRUNING_LEGR_RANDOM_SEED,
                        ),
                        "population_size": with_attributes(
                            NUMBER,
                            description="Size of population for the evolution algorithm.",
                            default=PRUNING_LEGR_POPULATION_SIZE,
                        ),
                        "num_samples": with_attributes(
                            NUMBER,
                            description="Number of samples for the evolution algorithm.",
                            default=PRUNING_LEGR_NUM_SAMPLES,
                        ),
                        "mutate_percent": with_attributes(
                            NUMBER,
                            description="Percent of mutate for the evolution algorithm.",
                            default=PRUNING_LEGR_MUTATE_PERCENT,
                        ),
                        "scale_sigma": with_attributes(
                            NUMBER,
                            description="Scale sigma for the evolution algorithm.",
                            default=PRUNING_LEGR_SIGMA_SCALE,
                        ),
                    },
                    "additionalProperties": False,
                },
                "save_ranking_coeffs_path": with_attributes(STRING),  # TODO(vshampor): is this important?
                "load_ranking_coeffs_path": with_attributes(STRING),  # TODO(vshampor): is this important?
            },
            "additionalProperties": False,
        },
        **SCOPING_PROPERTIES,
    },
    "additionalProperties": False,
}



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional, Tuple, TypeVar

import numpy as np

from nncf import Dataset
from nncf.common.factory import CommandCreatorFactory
from nncf.common.factory import ModelTransformerFactory
from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.graph import NNCFNode
from nncf.common.graph.patterns import GraphPattern
from nncf.common.graph.transformations.commands import TargetPoint
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.graph.transformations.layout import TransformationLayout
from nncf.common.graph.utils import get_reduction_axes
from nncf.common.logging import nncf_logger
from nncf.common.logging.track_progress import track
from nncf.common.tensor_statistics.statistic_point import StatisticPoint
from nncf.common.tensor_statistics.statistic_point import StatisticPointsContainer
from nncf.common.utils.backend import BackendType
from nncf.common.utils.backend import get_backend
from nncf.quantization.algorithms.algorithm import Algorithm
from nncf.quantization.algorithms.channel_alignment.backend import ChannelAlignmentAlgoBackend
from nncf.quantization.algorithms.channel_alignment.backend import LayoutDescriptor

TModel = TypeVar("TModel")


class ChannelAlignment(Algorithm):
    """
    Post-training ChannelAlignment algorithm implementation.

    The main purpose of this algorithm to reduce quantization error
    via correction the parameters of the Convolutions, FullyConnected and their biases.
    Algorithm consists of following steps:
        - algorithm is searching for convolution -> convolution pairs in the target model.
        - minimal and maximal activations quantiles of first convolutions are collected on the target subset.
        - algorithm calculates median of collected values, it is used then to adjust
            convolution layers biases and weights.
        - biases of matched subgraphs convolutions are adjusted, so mean points of first
            convolution activations quantile medians are translated to zero.
        - weights of matched subgraph convolutions are adjusted, so all first convolutions activations
            which were between median of low quantile and median of high quantile are translated to [-1, 1] range.
    In case processed network has one or more convolution -> convolution pairs, activations of the first convolution
    become more quantization friendly as, in most cases activations mean is equal to zero and
    most activations values are in range [-1, 1].
    """

    def __init__(
        self,
        subset_size: int = 100,
        inplace_statistics: bool = True,
    ):
        """
        :param subset_size: Size of a subset for the statistics collection,
            defaults to 100.
        :param inplace_statistics: Defines wheather to calculate quantizers statistics
            by backend graph operations or by default Python implementation, defaults
            to True.
        """
        super().__init__()
        self.subset_size = subset_size
        self.inplace_statistics = inplace_statistics
        self._backend_entity = None
        self._quantile = 1e-4
        self._algorithm_key = f"CA_{hash(self)}"

    @property
    def available_backends(self) -> List[BackendType]:
        return [BackendType.OPENVINO]

    def _set_backend_entity(self, model: TModel) -> None:
        """
        Creates a helper class with a backed-specific logic of the algorithm.

        :param model: Backend-specific input model.
        """
        model_backend = get_backend(model)
        if model_backend == BackendType.OPENVINO:
            from nncf.quantization.algorithms.channel_alignment.openvino_backend import OVChannelAlignmentAlgoBackend

            self._backend_entity = OVChannelAlignmentAlgoBackend()

    def apply(
        self,
        model: TModel,
        graph: NNCFGraph,
        statistic_points: Optional[StatisticPointsContainer] = None,
        dataset: Optional[Dataset] = None,
    ) -> TModel:
        self._set_backend_entity(model)
        model_transformer = ModelTransformerFactory.create(model)
        transformation_layout = TransformationLayout()

        def filter_func(point: StatisticPoint) -> bool:
            return self._algorithm_key in point.algorithm_to_tensor_collectors and point.target_point == target_point

        for conv_in, add_in, conv_out in track(self._get_node_pairs(graph), description="Channel alignment"):
            target_point, node_in = self._get_target_point_and_node_in(conv_in, add_in)
            tensor_collectors = list(
                statistic_points.get_algo_statistics_for_node(node_in.node_name, filter_func, self._algorithm_key)
            )
            assert len(tensor_collectors) == 1
            stat = tensor_collectors[0].get_statistics()
            if stat.min_values is None or stat.max_values is None:
                nncf_logger.debug(
                    f"Skipping channel alignment for pairs {conv_in.node_name}, {conv_out.node_name} "
                    "because statistics were not collected for this pair."
                )
                continue

            conv_in_cont = ConvParamsContainer(conv_in, model, graph, self._backend_entity)
            conv_out_cont = ConvParamsContainer(conv_out, model, graph, self._backend_entity)
            if (
                conv_in_cont.dims.conv_weight_out_channels_dim is None
                or conv_out_cont.dims.conv_weight_out_channels_dim is None
            ):
                nncf_logger.debug(
                    f"Skipping channel alignment for pairs {conv_in.node_name}, {conv_out.node_name} "
                    " because one of the node is 1D MatMul, 1D Matmuls are not supported by CA algortihm yet."
                )
                continue

            amean = (stat.max_values + stat.min_values) * 0.5
            conv_in_cont.bias, conv_out_cont.bias = self._align_means(
                conv_in_cont.bias,
                conv_out_cont.bias,
                conv_out_cont.weight,
                amean,
                conv_out_cont.dims,
            )

            ascale = (stat.max_values - stat.min_values).astype(np.float32)
            eps = np.finfo(ascale.dtype).eps
            if (ascale > eps).any():
                conv_in_cont.weight, conv_out_cont.weight, conv_in_cont.bias = self._align_scales(
                    conv_in_cont.weight,
                    conv_out_cont.weight,
                    conv_in_cont.bias,
                    ascale,
                    conv_in_cont.dims,
                    conv_out_cont.dims,
                    eps,
                )

            command_creator = CommandCreatorFactory.create(model)
            for container in [conv_in_cont, conv_out_cont]:
                if container.stated_weight.is_modified():
                    transformation_layout.register(
                        command_creator.create_command_to_update_weight(
                            container.op, container.weight, container.weight_port_id
                        )
                    )

                if container.stated_bias.is_modified():
                    if container.bias_op_exist():
                        command = command_creator.create_command_to_update_bias(container.op, container.bias, graph)
                    else:
                        command = command_creator.create_command_to_insert_bias(container.op, container.bias)
                    transformation_layout.register(command)

        transformed_model = model_transformer.transform(transformation_layout)
        return transformed_model

    @staticmethod
    def _align_means(
        bias_in_value: np.ndarray,
        bias_out_value: np.ndarray,
        conv_out_value: np.ndarray,
        amean: np.ndarray,
        conv_out_descr: LayoutDescriptor,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Function which calculates new add_in_value and add_out_value
        in ChannelAlignment pattern, so output activations of the second convolution bias
        are the same, but the first convolution bias is shifted with minus by amean value.

        :param bias_in_value: Bias of the first convolution in the ChannelAlignment pattern.
        :param bias_out_value: Bias of the second convolution in the ChannelAlignment pattern.
        :param amean: Mean value to shift first and second convolutions biases.
        :param conv_out_descr: The second convolution weights layout descriptor.
        """
        updated_add_in_value = bias_in_value - amean.reshape(bias_in_value.shape)

        weight_dims = conv_out_value.ndim
        updated_conv_out_value = conv_out_value
        if weight_dims > 2:
            axes = list(range(weight_dims))
            axes.remove(conv_out_descr.conv_weight_in_channels_dim)
            axes.remove(conv_out_descr.conv_weight_out_channels_dim)
            updated_conv_out_value = np.sum(conv_out_value, axis=tuple(axes))

        out_channel_dim, in_channel_dim = 0, 1
        if conv_out_descr.conv_weight_out_channels_dim > conv_out_descr.conv_weight_in_channels_dim:
            out_channel_dim, in_channel_dim = in_channel_dim, out_channel_dim

        updated_conv_out_value = np.transpose(
            updated_conv_out_value,
            (out_channel_dim, in_channel_dim),
        )
        shift = updated_conv_out_value.dot(amean.reshape(updated_conv_out_value.shape[1]))

        updated_add_out_value = bias_out_value + shift.reshape(bias_out_value.shape)
        return updated_add_in_value, updated_add_out_value

    @staticmethod
    def _align_scales(
        conv_in_value: np.ndarray,
        conv_out_value: np.ndarray,
        bias_in_value: Optional[np.ndarray],
        ascale: np.ndarray,
        conv_in_descr: LayoutDescriptor,
        conv_out_descr: LayoutDescriptor,
        eps: float,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Function which calculates new conv_in_value, conv_out_value and bias_in_value
        in ChannelAlignment pattern, so output activations of conv_out are the same,
        but activations of conv_in are scale times smaller. Negative scales are skipped,
        too small (<1e-2) and too big (>1e2) scales are clamped.

        :param conv_in_value: Weights of the first convolution in the ChannelAlignment pattern.
        :param conv_out_value: Weights of the second convolution in the ChannelAlignment pattern.
        :param bias_in_value: Bias of the first convolution in the ChannelAlignment pattern. Could be None.
        :param ascale: Scale value to apply to convolutions weights.
        :param conv_in_descr: The first convolution weights layout descriptor.
        :param conv_out_descr: The second convolution weights layout descriptor.
        :param eps: Minimal significant value > 0 for convolution weights and biases precision.
        """
        conv_in_shape = conv_in_value.shape
        # TODO(dlyakhov) support group convolutions with groups number not in [1, out_channels]
        if conv_in_shape[conv_in_descr.conv_weight_out_channels_dim] != ascale.shape[conv_in_descr.bias_channels_dim]:
            return conv_in_value, conv_out_value, bias_in_value

        positive_scales_mask = ascale > eps
        scale_factor = ascale / np.median(ascale[positive_scales_mask])
        scale_factor[~positive_scales_mask] = 1
        scale_factor = np.clip(scale_factor, 1e-2, 1e2)

        scale_in_shape = np.ones(len(conv_in_shape), dtype=int)
        scale_in_shape[conv_in_descr.conv_weight_out_channels_dim] = scale_factor.shape[conv_in_descr.bias_channels_dim]
        updated_conv_in_value = conv_in_value / scale_factor.reshape(scale_in_shape)

        updated_bias_in_value = bias_in_value / scale_factor.reshape(bias_in_value.shape)

        scale_out_shape = np.ones(len(conv_out_value.shape), dtype=int)
        scale_out_shape[conv_out_descr.conv_weight_in_channels_dim] = scale_factor.shape[
            conv_in_descr.bias_channels_dim
        ]
        updated_conv_out_value = conv_out_value * scale_factor.reshape(scale_out_shape)
        return updated_conv_in_value, updated_conv_out_value, updated_bias_in_value

    def _check_consumer_conv_node(self, conv_node: NNCFNode) -> bool:
        attrs = self._backend_entity.get_conv_layer_attributes(conv_node)
        if attrs is None:
            return False
        # Check groups amount == 1
        if attrs.groups != 1:
            return False
        # Check node has no padding
        if any(attrs.padding_values):
            return False
        # Check node has valid stride
        if any(elem != 1 for elem in attrs.stride):
            return False
        # Check Node has valid dilation
        if any(elem != 1 for elem in attrs.dilations):
            return False
        return True

    def _check_producer_conv_node(self, conv_node: NNCFNode):
        return conv_node.layer_attributes is not None

    def _get_target_patterns(self) -> GraphPattern:
        input_attrs = {
            GraphPattern.LABEL_ATTR: "INPUT",
            GraphPattern.METATYPE_ATTR: GraphPattern.NON_PATTERN_NODE_TYPE,
        }
        producer_attrs = {
            GraphPattern.LABEL_ATTR: "CONV_PRODUCER",
            GraphPattern.NODE_TYPE_ATTR: self._backend_entity.get_conv_metatypes()
            + self._backend_entity.get_linear_metatypes(),
        }
        bias_attrs = {
            GraphPattern.LABEL_ATTR: "BIAS_PRODUCER",
            GraphPattern.NODE_TYPE_ATTR: self._backend_entity.get_add_metatypes(),
        }
        bias_const_attrs = {
            GraphPattern.LABEL_ATTR: "BIAS_CONSTANT",
            GraphPattern.METATYPE_ATTR: GraphPattern.NON_PATTERN_NODE_TYPE,
        }
        consumer_attrs = {
            GraphPattern.LABEL_ATTR: "CONV_CONSUMER",
            GraphPattern.NODE_TYPE_ATTR: self._backend_entity.get_conv_metatypes(),
        }
        conv_const_attrs = {
            GraphPattern.LABEL_ATTR: "CONV_CONSTANT",
            GraphPattern.METATYPE_ATTR: GraphPattern.NON_PATTERN_NODE_TYPE,
        }

        use_constant = True

        def get_conv_conv_pattern() -> GraphPattern:
            conv_conv = GraphPattern()
            if use_constant:
                input_node = conv_conv.add_node(**input_attrs)
                producer_constant = conv_conv.add_node(**conv_const_attrs)
                consumer_constant = conv_conv.add_node(**conv_const_attrs)

            pattern_conv_producer = conv_conv.add_node(**producer_attrs)
            pattern_conv_consumer = conv_conv.add_node(**consumer_attrs)

            if use_constant:
                conv_conv.add_edge(input_node, pattern_conv_producer)
                conv_conv.add_edge(producer_constant, pattern_conv_producer)
                conv_conv.add_edge(consumer_constant, pattern_conv_consumer)

            conv_conv.add_edge(pattern_conv_producer, pattern_conv_consumer)
            return conv_conv

        def get_conv_add_conv_pattern() -> GraphPattern:
            conv_bias_conv = GraphPattern()
            if use_constant:
                input_node = conv_bias_conv.add_node(**input_attrs)
                producer_constant = conv_bias_conv.add_node(**conv_const_attrs)
                bias_producer_const = conv_bias_conv.add_node(**bias_const_attrs)
                consumer_constant = conv_bias_conv.add_node(**conv_const_attrs)

            pattern_conv_producer = conv_bias_conv.add_node(**producer_attrs)
            pattern_bias_producer = conv_bias_conv.add_node(**bias_attrs)
            pattern_conv_consumer = conv_bias_conv.add_node(**consumer_attrs)

            if use_constant:
                conv_bias_conv.add_edge(input_node, pattern_conv_producer)
                conv_bias_conv.add_edge(producer_constant, pattern_conv_producer)
                conv_bias_conv.add_edge(consumer_constant, pattern_conv_consumer)
                conv_bias_conv.add_edge(bias_producer_const, pattern_bias_producer)

            conv_bias_conv.add_edge(pattern_conv_producer, pattern_bias_producer)
            conv_bias_conv.add_edge(pattern_bias_producer, pattern_conv_consumer)
            return conv_bias_conv

        pattern = get_conv_conv_pattern()
        pattern.add_pattern_alternative(get_conv_add_conv_pattern())
        return pattern

    def _get_node_pairs(self, nncf_graph: NNCFGraph) -> List[Tuple[NNCFNode, Optional[NNCFNode], NNCFNode]]:
        pairs = []
        patterns = self._get_target_patterns()
        for subgraph in nncf_graph.find_matching_subgraphs(patterns):
            if len(subgraph) == 2:
                add_in = None
                conv_in, conv_out = subgraph
            else:
                conv_in, add_in, conv_out = subgraph

            if not self._check_producer_conv_node(conv_in):
                continue

            if not self._check_consumer_conv_node(conv_out):
                continue

            pairs.append((conv_in, add_in, conv_out))
        return pairs

    def _get_target_point_and_node_in(self, conv_in, add_in) -> Tuple[TargetPoint, NNCFNode]:
        node_in = conv_in if add_in is None else add_in
        input_port_id, _ = self._backend_entity.get_activation_port_ids_for_node(node_in)
        return (
            self._backend_entity.target_point(TargetType.POST_LAYER_OPERATION, node_in.node_name, input_port_id),
            node_in,
        )

    def get_statistic_points(self, model: TModel, graph: NNCFGraph) -> StatisticPointsContainer:
        self._set_backend_entity(model)

        statistic_container = StatisticPointsContainer()
        for conv_in, add_in, _ in self._get_node_pairs(graph):
            target_point, node_in = self._get_target_point_and_node_in(conv_in, add_in)

            channel_axis = conv_in.metatype.output_channel_axis
            activation_shape = list(range(len(graph.get_output_edges(node_in)[0].tensor_shape)))
            reduction_axes = get_reduction_axes([0, channel_axis], activation_shape)

            statistic_collector = self._backend_entity.get_statistic_collector(
                reduction_axes, self._quantile, self.subset_size, self.inplace_statistics
            )
            statistic_container.add_statistic_point(
                StatisticPoint(
                    target_point=target_point,
                    tensor_collector=statistic_collector,
                    algorithm=self._algorithm_key,
                )
            )

        return statistic_container


class StatedTensor:
    """
    Tensor wrapper with additional method is_modified which is true if
    given tensor was modified at least once after the initialization.
    """

    def __init__(self, value: np.ndarray):
        """
        :param value: Tensor to wrap.
        """
        self._value = value
        self._mod_times = 0

    @property
    def val(self):
        return self._value

    @val.setter
    def val(self, value):
        if self._value is None and value is None:
            return
        self._mod_times += 1
        self._value = value

    def is_modified(self) -> bool:
        """
        :return: True if wrapped tensor was changed at least once after the
            initialization else False.
        """
        return self._mod_times > 0


class ConvParamsContainer:
    """
    Convolution container class which is incapsulating common convolutional parameters collection.
    """

    def __init__(
        self, conv_op: NNCFNode, model: TModel, nncf_graph: NNCFGraph, backend_entity: ChannelAlignmentAlgoBackend
    ):
        """
        :param conv_op: NNCF conv node.
        :param model: Backend-specific model instance.
        :param nncf_graph: NNCFGraph of given backend-specific model.
        :param backend_entity: Current backend entity to retrieve parameters from given conv node
        """
        _, self._weights_port_id = backend_entity.get_weights_port_ids_for_node(conv_op)
        self.stated_weight = StatedTensor(backend_entity.get_weight_value(conv_op, model, self._weights_port_id))
        self._bias_op_exist = False
        if backend_entity.is_node_with_bias(conv_op, nncf_graph):
            bias = backend_entity.get_bias_value(conv_op, model, nncf_graph)
            self._bias_op_exist = True
        else:
            bias = backend_entity.create_bias_tensor(conv_op, nncf_graph, 0)
        self.stated_bias = StatedTensor(bias)
        self._op = conv_op
        self._dims = backend_entity.get_dims_descriptor(conv_op)

    @property
    def weight(self):
        return self.stated_weight.val

    @weight.setter
    def weight(self, value):
        self.stated_weight.val = value

    @property
    def bias(self):
        return self.stated_bias.val

    @bias.setter
    def bias(self, value):
        self.stated_bias.val = value

    @property
    def op(self):
        return self._op

    @property
    def weight_port_id(self):
        return self._weights_port_id

    @property
    def dims(self) -> LayoutDescriptor:
        return self._dims

    def bias_op_exist(self) -> bool:
        return self._bias_op_exist



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import types
from typing import Any, Dict, Optional, Tuple

import tensorflow as tf

import nncf
from nncf import NNCFConfig
from nncf.api.compression import CompressionAlgorithmController
from nncf.common.compression import BaseCompressionAlgorithmController as BaseController
from nncf.common.utils.api_marker import api
from nncf.config.extractors import extract_algorithm_names
from nncf.config.telemetry_extractors import CompressionStartedFromConfig
from nncf.config.utils import is_experimental_quantization
from nncf.telemetry import tracked_function
from nncf.telemetry.events import NNCF_TF_CATEGORY
from nncf.tensorflow.accuracy_aware_training.keras_model_utils import accuracy_aware_fit
from nncf.tensorflow.algorithm_selector import NoCompressionAlgorithmBuilder
from nncf.tensorflow.algorithm_selector import get_compression_algorithm_builder
from nncf.tensorflow.api.composite_compression import TFCompositeCompressionAlgorithmBuilder
from nncf.tensorflow.api.compression import TFCompressionAlgorithmBuilder
from nncf.tensorflow.graph.utils import is_keras_layer_model
from nncf.tensorflow.helpers.utils import get_built_model


def create_compression_algorithm_builder(config: NNCFConfig, should_init: bool) -> TFCompressionAlgorithmBuilder:
    """
    Factory to create an instance of the compression algorithm builder
    by NNCFConfig.

    :param config: An instance of NNCFConfig that defines compression methods.
    :param should_init: The flag indicates that the generated compression builder
        will initialize (True) or not (False) the training parameters of the model
        during model building.
    :return: An instance of the `CompressionAlgorithmBuilder`
    """
    algo_names = extract_algorithm_names(config)
    number_compression_algorithms = len(algo_names)
    if number_compression_algorithms == 0:
        return NoCompressionAlgorithmBuilder(config, should_init)
    if number_compression_algorithms == 1:
        algo_name = next(iter(algo_names))
        return get_compression_algorithm_builder(algo_name)(config, should_init)

    return TFCompositeCompressionAlgorithmBuilder(config, should_init)


@api(canonical_alias="nncf.tensorflow.create_compressed_model")
@tracked_function(
    NNCF_TF_CATEGORY,
    [
        CompressionStartedFromConfig(argname="config"),
    ],
)
def create_compressed_model(
    model: tf.keras.Model, config: NNCFConfig, compression_state: Optional[Dict[str, Any]] = None
) -> Tuple[CompressionAlgorithmController, tf.keras.Model]:
    """
    The main function used to produce a model ready for compression fine-tuning
    from an original TensorFlow Keras model and a configuration object.

    :param model: The original model. Should have its parameters already loaded
        from a checkpoint or another source.
    :param config: A configuration object used to determine the exact compression
        modifications to be applied to the model.
    :type config: nncf.NNCFConfig
    :param compression_state: compression state to unambiguously restore the compressed model.
        Includes builder and controller states. If it is specified, trainable parameter initialization will be skipped
        during building.
    :return: A tuple of the compression controller for the requested algorithm(s) and the model object with additional
     modifications necessary to enable algorithm-specific compression during fine-tuning.
    """
    if is_experimental_quantization(config):
        if is_keras_layer_model(model):
            raise ValueError(
                "Experimental quantization algorithm has not supported models with "
                "`tensorflow_hub.KerasLayer` layer yet."
            )

        from nncf.experimental.tensorflow.nncf_network import NNCFNetwork

        input_signature = get_input_signature(config)
        model = NNCFNetwork(model, input_signature)
        model.compute_output_signature(model.input_signature)

    model = get_built_model(model, config)

    builder = create_compression_algorithm_builder(config, should_init=not compression_state)

    if compression_state:
        builder.load_state(compression_state[BaseController.BUILDER_STATE])
    compressed_model = builder.apply_to(model)
    compression_ctrl = builder.build_controller(compressed_model)
    if isinstance(compressed_model, tf.keras.Model):
        compressed_model.accuracy_aware_fit = types.MethodType(accuracy_aware_fit, compressed_model)
    return compression_ctrl, compressed_model


def get_input_signature(config: NNCFConfig):
    input_info = config.get("input_info", {})
    samples_sizes = []

    if isinstance(input_info, dict):
        sample_size = input_info["sample_size"]
        samples_sizes.append(sample_size)
    elif isinstance(input_info, list):
        for info in input_info:
            sample_size = info["sample_size"]
            samples_sizes.append(sample_size)
    else:
        raise nncf.ValidationError("sample_size must be provided in configuration file")

    input_signature = []
    for sample_size in samples_sizes:
        shape = [None] + list(sample_size[1:])
        input_signature.append(tf.TensorSpec(shape=shape, dtype=tf.float32))

    return input_signature if len(input_signature) > 1 else input_signature[0]



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict

import torch

from nncf.api.compression import CompressionScheduler
from nncf.api.compression import CompressionStage
from nncf.common.compression import NO_COMPRESSION_ALGORITHM_NAME
from nncf.common.schedulers import StubCompressionScheduler
from nncf.common.statistics import NNCFStatistics
from nncf.common.utils.backend import copy_model
from nncf.common.utils.registry import Registry
from nncf.torch.compression_method_api import PTCompressionAlgorithmBuilder
from nncf.torch.compression_method_api import PTCompressionAlgorithmController
from nncf.torch.compression_method_api import PTCompressionLoss
from nncf.torch.graph.transformations.layout import PTTransformationLayout
from nncf.torch.nncf_network import NNCFNetwork
from nncf.torch.utils import get_model_device

PT_COMPRESSION_ALGORITHMS = Registry("compression algorithm", add_name_as_attr=True)


class ZeroCompressionLoss(PTCompressionLoss):
    def __init__(self, device: torch.device):
        super().__init__()
        self._device = device

    def calculate(self) -> torch.Tensor:
        return torch.zeros([], device=self._device)


@PT_COMPRESSION_ALGORITHMS.register(NO_COMPRESSION_ALGORITHM_NAME)
class NoCompressionAlgorithmBuilder(PTCompressionAlgorithmBuilder):
    def _get_transformation_layout(self, target_model: NNCFNetwork) -> PTTransformationLayout:
        return PTTransformationLayout()

    def _get_algo_specific_config_section(self) -> Dict:
        return {}

    def _build_controller(self, model: NNCFNetwork) -> PTCompressionAlgorithmController:
        return NoCompressionAlgorithmController(model)

    def initialize(self, model: NNCFNetwork) -> None:
        pass

    def _handle_frozen_layers(self, target_model: NNCFNetwork):
        pass


class NoCompressionAlgorithmController(PTCompressionAlgorithmController):
    def __init__(self, target_model):
        super().__init__(target_model)

        self._loss = ZeroCompressionLoss(get_model_device(target_model))
        self._scheduler = StubCompressionScheduler()

    def compression_stage(self) -> CompressionStage:
        """
        Returns level of compression. Should be used on saving best checkpoints to distinguish between
        uncompressed, partially compressed and fully compressed models.
        """
        return CompressionStage.UNCOMPRESSED

    @property
    def loss(self) -> ZeroCompressionLoss:
        return self._loss

    @property
    def scheduler(self) -> CompressionScheduler:
        return self._scheduler

    def statistics(self, quickly_collected_only: bool = False) -> NNCFStatistics:
        return NNCFStatistics()

    def strip(self, do_copy: bool = True) -> NNCFNetwork:
        model = self.model
        if do_copy:
            model = copy_model(self.model)
        return model



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf


def symmetric_quantize(
    inputs, scale_var, signed_var, num_bits, per_channel, narrow_range, eps, name_prefix="SymmQuant"
):
    with tf.name_scope(name_prefix):
        scale_safe = tf.abs(scale_var) + eps
        min_var = scale_safe * signed_var
        max_var = scale_safe
        return _fake_quant_with_min_max_vars(inputs, min_var, max_var, num_bits, narrow_range, per_channel)


def asymmetric_quantize(
    inputs, input_low, input_range, num_bits, per_channel, narrow_range, eps, name_prefix="AsymmQuant"
):
    with tf.name_scope(name_prefix):
        input_range_safe = tf.abs(input_range) + eps
        min_var = input_low
        max_var = input_low + input_range_safe
        return _fake_quant_with_min_max_vars(inputs, min_var, max_var, num_bits, narrow_range, per_channel)


def _fake_quant_with_min_max_vars(inputs, min_var, max_var, num_bits, narrow_range, per_channel):
    if per_channel:
        return tf.quantization.fake_quant_with_min_max_vars_per_channel(
            inputs, min_var, max_var, num_bits=num_bits, narrow_range=narrow_range
        )
    return tf.quantization.fake_quant_with_min_max_vars(
        inputs, min_var, max_var, num_bits=num_bits, narrow_range=narrow_range
    )



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Iterable, List, Optional, Sequence, Union

from rich.console import Console
from rich.progress import BarColumn
from rich.progress import Column
from rich.progress import Progress
from rich.progress import ProgressColumn
from rich.progress import ProgressType
from rich.progress import Task
from rich.progress import TaskProgressColumn
from rich.progress import TextColumn
from rich.progress import TimeElapsedColumn
from rich.progress import TimeRemainingColumn
from rich.style import StyleType
from rich.text import Text

INTEL_BLUE_COLOR = "#0068b5"


class IterationsColumn(ProgressColumn):
    def render(self, task: Task) -> Text:
        if task.total is None:
            return Text("")
        text = f"{int(task.completed)}/{int(task.total)}"
        return Text(text, style=INTEL_BLUE_COLOR)


class SeparatorColumn(ProgressColumn):
    def __init__(self, table_column: Optional[Column] = None, disable_if_no_total: bool = False) -> None:
        super().__init__(table_column)
        self.disable_if_no_total = disable_if_no_total

    def render(self, task: Task) -> Text:
        if self.disable_if_no_total and task.total is None:
            return Text("")
        return Text("•")


class TimeElapsedColumnWithStyle(TimeElapsedColumn):
    def render(self, task: "Task") -> Text:
        text = super().render(task)
        return Text(text._text[0], style=INTEL_BLUE_COLOR)


class TimeRemainingColumnWithStyle(TimeRemainingColumn):
    def render(self, task: "Task") -> Text:
        text = super().render(task)
        return Text(text._text[0], style=INTEL_BLUE_COLOR)


class track:
    def __init__(
        self,
        sequence: Optional[Union[Sequence[ProgressType], Iterable[ProgressType]]] = None,
        description: str = "Working...",
        total: Optional[float] = None,
        auto_refresh: bool = True,
        console: Optional[Console] = None,
        transient: bool = False,
        get_time: Optional[Callable[[], float]] = None,
        refresh_per_second: float = 10,
        style: StyleType = "bar.back",
        complete_style: StyleType = "bar.complete",
        finished_style: StyleType = "bar.finished",
        pulse_style: StyleType = "bar.pulse",
        update_period: float = 0.1,
        disable: bool = False,
        show_speed: bool = True,
    ):
        """
        Track progress by iterating over a sequence.

        This function is very similar to rich.progress.track(), but with some customizations.

        :param sequence: An iterable (must support "len") you wish to iterate over.
        :param description: Description of the task to show next to the progress bar. Defaults to "Working".
        :param total: Total number of steps. Default is len(sequence).
        :param auto_refresh: Automatic refresh. Disable to force a refresh after each iteration. Default is True.
        :param transient: Clear the progress on exit. Defaults to False.
        :param get_time: A callable that gets the current time, or None to use Console.get_time. Defaults to None.
        :param console: Console to write to. Default creates an internal Console instance.
        :param refresh_per_second: Number of times per second to refresh the progress information. Defaults to 10.
        :param style: Style for the bar background. Defaults to "bar.back".
        :param complete_style: Style for the completed bar. Defaults to "bar.complete".
        :param finished_style: Style for a finished bar. Defaults to "bar.finished".
        :param pulse_style: Style for pulsing bars. Defaults to "bar.pulse".
        :param update_period: Minimum time (in seconds) between calls to update(). Defaults to 0.1.
        :param disable: Disable display of progress.
        :param show_speed: Show speed if the total isn't known. Defaults to True.
        :return: An iterable of the values in the sequence.
        """

        self.sequence = sequence
        self.total = total
        self.description = description
        self.update_period = update_period
        self.task = None

        self.columns: List[ProgressColumn] = (
            [TextColumn("[progress.description]{task.description}")] if description else []
        )
        self.columns.extend(
            (
                BarColumn(
                    style=style,
                    complete_style=complete_style,
                    finished_style=finished_style,
                    pulse_style=pulse_style,
                    bar_width=None,
                ),
                TaskProgressColumn(show_speed=show_speed),
                IterationsColumn(),
                SeparatorColumn(),
                TimeElapsedColumnWithStyle(),
                SeparatorColumn(disable_if_no_total=True),  # disable because time remaining will be empty
                TimeRemainingColumnWithStyle(),
            )
        )

        disable = disable or (hasattr(sequence, "__len__") and len(sequence) == 0)

        self.progress = Progress(
            *self.columns,
            auto_refresh=auto_refresh,
            console=console,
            transient=transient,
            get_time=get_time,
            refresh_per_second=refresh_per_second or 10,
            disable=disable,
        )

    def __iter__(self) -> Iterable[ProgressType]:
        with self.progress:
            yield from self.progress.track(
                self.sequence, total=self.total, description=self.description, update_period=self.update_period
            )

    def __enter__(self):
        self.progress.start()
        self.task = self.progress.add_task(self.description, total=self.total)
        return self

    def __exit__(self, *args):
        self.task = None
        self.progress.stop()
